from datetime import date
from pathlib import Path
from typing import Annotated, Any

import questionary
import typer
from cognite.client.data_classes.data_modeling import ContainerId

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands import (
    MigrationCanvasCommand,
    MigrationPrepareCommand,
)
from cognite_toolkit._cdf_tk.commands._migrate import MigrationCommand
from cognite_toolkit._cdf_tk.commands._migrate.creators import (
    InfieldV2ConfigCreator,
    InstanceSpaceCreator,
    SourceSystemCreator,
)
from cognite_toolkit._cdf_tk.commands._migrate.data_mapper import AssetCentricMapper
from cognite_toolkit._cdf_tk.commands._migrate.migration_io import (
    AssetCentricMigrationIO,
)
from cognite_toolkit._cdf_tk.commands._migrate.selectors import (
    AssetCentricMigrationSelector,
    MigrateDataSetSelector,
    MigrationCSVFileSelector,
)
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from cognite_toolkit._cdf_tk.utils.cli_args import parse_view_str
from cognite_toolkit._cdf_tk.utils.interactive_select import (
    AssetInteractiveSelect,
    DataModelingSelect,
    ResourceViewMappingInteractiveSelect,
)
from cognite_toolkit._cdf_tk.utils.useful_types import AssetCentricKind

TODAY = date.today()


class MigrateApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.command("prepare")(self.prepare)
        self.command("data-sets")(self.data_sets)
        self.command("source-systems")(self.source_systems)
        self.command("assets")(self.assets)
        self.command("events")(self.events)
        self.command("timeseries")(self.timeseries)
        self.command("files")(self.files)
        self.command("canvas")(self.canvas)
        # Uncomment when infield v2 config migration is ready
        # self.command("infield-configs")(self.infield_configs)

    def main(self, ctx: typer.Context) -> None:
        """Migrate resources from Asset-Centric to data modeling in CDF."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf migrate --help[/] for more information.")

    @staticmethod
    def prepare(
        ctx: typer.Context,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="The preparation will not be executed, only report of what would be done is printed.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Prepare the migration of resources from Asset-Centric to data modeling in CDF.

        This means deploying the CogniteMigration data model that contains the Mapping view. This will be used
        to store the mapping from Asset-Centric resources to the new data modeling resources.

        This mapping will be used when migrating applications such as Canvas, Charts, as well as resources that
        depend on the primary resources 3D and annotations.
        """
        client = EnvironmentVariables.create_from_environment().get_client(enable_set_pending_ids=True)
        cmd = MigrationPrepareCommand()
        cmd.run(
            lambda: cmd.deploy_cognite_migration(
                client,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @staticmethod
    def data_sets(
        ctx: typer.Context,
        data_set: Annotated[
            list[str] | None,
            typer.Argument(
                help="The name or external ID of the data set to create Instance Spaces for. If not provided, an "
                "interactive selection will be performed to select the data sets to create Instance Spaces for."
            ),
        ] = None,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Path to the directory where the instance space definitions will be dumped. It is recommended "
                "to govern these configurations in a git repository.",
            ),
        ] = Path("tmp"),
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Creates Instance Spaces for all selected data sets."""
        client = EnvironmentVariables.create_from_environment().get_client()
        if data_set is None:
            # Interactive model
            selector = AssetInteractiveSelect(client, "migrate")
            data_set = selector.select_data_sets()
            dry_run = questionary.confirm("Do you want to perform a dry run?", default=dry_run).ask()
            output_dir = questionary.path(
                "Specify output directory for instance space definitions:", default=str(output_dir)
            ).ask()
            verbose = questionary.confirm("Do you want verbose output?", default=verbose).ask()
            if any(res is None for res in [dry_run, output_dir, verbose]):
                raise typer.Abort()
            output_dir = Path(output_dir)

        cmd = MigrationCommand()
        cmd.run(
            lambda: cmd.create(
                client,
                creator=InstanceSpaceCreator(client, data_set_external_ids=data_set),
                output_dir=output_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @staticmethod
    def source_systems(
        ctx: typer.Context,
        data_set: Annotated[
            str | None,
            typer.Argument(
                help="The external ID of the data set to lookup source system for. If not provided, an interactive "
                "selection will be performed to select the data sets to create instance spaces for."
            ),
        ] = None,
        instance_space: Annotated[
            str | None,
            typer.Option(
                "--instance-space",
                "-s",
                help="The instance space were you want to create the source system.",
            ),
        ] = None,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Path to the directory where the instance space definitions will be dumped. It is recommended "
                "to govern these configurations in a git repository.",
            ),
        ] = Path("tmp"),
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Creates source systems from the 'source' property of classic resources (assets, events, files)."""
        client = EnvironmentVariables.create_from_environment().get_client()
        if data_set is None and instance_space is None:
            # Interactive model
            ds_selector = AssetInteractiveSelect(client, "migrate")
            data_set = ds_selector.select_data_set()
            dm_selector = DataModelingSelect(client, "migrate")
            instance_space = dm_selector.select_instance_space(
                multiselect=False,
                message="In which instance space do you want to create the source system?",
                include_empty=True,
            )
            dry_run = questionary.confirm("Do you want to perform a dry run?", default=dry_run).ask()
            output_dir = questionary.path(
                "Specify output directory for instance space definitions:", default=str(output_dir)
            ).ask()
            verbose = questionary.confirm("Do you want verbose output?", default=verbose).ask()
            if any(res is None for res in [instance_space, dry_run, output_dir, verbose]):
                raise typer.Abort()
            output_dir = Path(output_dir)
        elif data_set is None or instance_space is None:
            raise typer.BadParameter("Both data_set and instance_space must be provided together.")

        cmd = MigrationCommand()
        cmd.run(
            lambda: cmd.create(
                client,
                creator=SourceSystemCreator(client, data_set_external_id=data_set, instance_space=instance_space),
                output_dir=output_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @classmethod
    def assets(
        cls,
        ctx: typer.Context,
        mapping_file: Annotated[
            Path | None,
            typer.Option(
                "--mapping-file",
                "-m",
                help="Path to the mapping file that contains the mapping from Assets to CogniteAssets. "
                "This file is expected to have the following columns: [id, dataSetId, space, externalId]."
                "The dataSetId is optional, and can be skipped. If it is set, it is used to check the access to the dataset.",
            ),
        ] = None,
        data_set_id: Annotated[
            str | None,
            typer.Option(
                "--data-set-id",
                "-s",
                help="The data set ID to use for the migrated CogniteAssets. If not provided, the dataSetId from the mapping file is used. "
                "If neither is provided, the default data set for the project is used.",
            ),
        ] = None,
        ingestion_mapping: Annotated[
            str | None,
            typer.Option(
                "--ingestion-mapping",
                "-i",
                help="The ingestion mapping to use for the migrated assets. If not provided, "
                "the default mapping to CogniteAsset in CogniteCore will be used.",
            ),
        ] = None,
        consumption_view: Annotated[
            str | None,
            typer.Option(
                "--consumption-view",
                "-c",
                help="The consumption view to assign to the migrated assets Given as space:externalId/version. "
                "This will be used in Canvas to select which view to use when migrating assets. If not provided, "
                "CogniteAsset in CogniteCore will be used.",
            ),
        ] = None,
        log_dir: Annotated[
            Path,
            typer.Option(
                "--log-dir",
                "-l",
                help="Path to the directory where logs will be stored. If the directory does not exist, it will be created.",
            ),
        ] = Path(f"migration_logs_{TODAY!s}"),
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Migrate Assets to CogniteAssets."""
        client = EnvironmentVariables.create_from_environment().get_client()
        selected, dry_run, verbose = cls._prepare_asset_centric_arguments(
            client=client,
            mapping_file=mapping_file,
            data_set_id=data_set_id,
            consumption_view=consumption_view,
            ingestion_mapping=ingestion_mapping,
            dry_run=dry_run,
            verbose=verbose,
            kind="Assets",
            resource_type="asset",
            container_id=ContainerId("cdf_cdm", "CogniteAsset"),
        )

        cmd = MigrationCommand()
        cmd.run(
            lambda: cmd.migrate(
                selected=selected,
                data=AssetCentricMigrationIO(client),
                mapper=AssetCentricMapper(client),
                log_dir=log_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @staticmethod
    def _prepare_asset_centric_arguments(
        client: ToolkitClient,
        mapping_file: Path | None,
        data_set_id: str | None,
        consumption_view: str | None,
        ingestion_mapping: str | None,
        dry_run: bool,
        verbose: bool,
        kind: AssetCentricKind,
        resource_type: str,
        container_id: ContainerId,
    ) -> tuple[AssetCentricMigrationSelector, bool, bool]:
        if data_set_id is not None and mapping_file is not None:
            raise typer.BadParameter("Cannot specify both data_set_id and mapping_file")
        elif mapping_file is not None:
            selected: AssetCentricMigrationSelector = MigrationCSVFileSelector(datafile=mapping_file, kind=kind)
        elif data_set_id is not None:
            parsed_view = parse_view_str(consumption_view) if consumption_view is not None else None
            selected = MigrateDataSetSelector(
                data_set_external_id=data_set_id,
                kind=kind,
                ingestion_mapping=ingestion_mapping,
                preferred_consumer_view=parsed_view,
            )
        else:
            # Interactive selection of data set.
            selector = AssetInteractiveSelect(client, "migrate")
            selected_data_set_id = selector.select_data_set(allow_empty=False)
            asset_mapping = ResourceViewMappingInteractiveSelect(client, "migrate").select_resource_view_mapping(
                resource_type,
            )
            preferred_consumer_view = (
                DataModelingSelect(client, "migrate")
                .select_view(
                    multiselect=False,
                    include_global=True,
                    instance_type="node",
                    mapped_container=container_id,
                )
                .as_id()
            )
            selected = MigrateDataSetSelector(
                data_set_external_id=selected_data_set_id,
                kind=kind,
                ingestion_mapping=asset_mapping.external_id,
                preferred_consumer_view=preferred_consumer_view,
            )
            dry_run = questionary.confirm("Do you want to perform a dry run?", default=dry_run).ask()
            verbose = questionary.confirm("Do you want verbose output?", default=verbose).ask()
            if any(res is None for res in [dry_run, verbose]):
                raise typer.Abort()
        return selected, dry_run, verbose

    @classmethod
    def events(
        cls,
        ctx: typer.Context,
        mapping_file: Annotated[
            Path | None,
            typer.Option(
                "--mapping-file",
                "-m",
                help="Path to the mapping file that contains the mapping from Events to CogniteActivity. "
                "This file is expected to have the following columns: [id, dataSetId, space, externalId]."
                "The dataSetId is optional, and can be skipped. If it is set, it is used to check the access to the dataset.",
            ),
        ] = None,
        data_set_id: Annotated[
            str | None,
            typer.Option(
                "--data-set-id",
                "-s",
                help="The data set ID to use for the migrated CogniteActivity. If not provided, the dataSetId from the mapping file is used. "
                "If neither is provided, the default data set for the project is used.",
            ),
        ] = None,
        ingestion_mapping: Annotated[
            str | None,
            typer.Option(
                "--ingestion-mapping",
                "-i",
                help="The ingestion mapping to use for the migrated events. If not provided, "
                "the default mapping to CogniteActivity in CogniteCore will be used.",
            ),
        ] = None,
        consumption_view: Annotated[
            str | None,
            typer.Option(
                "--consumption-view",
                "-c",
                help="The consumption view to assign to the migrated events Given as space:externalId/version. "
                "This will be used in Canvas to select which view to use when migrating events. If not provided, "
                "CogniteActivity in CogniteCore will be used.",
            ),
        ] = None,
        log_dir: Annotated[
            Path,
            typer.Option(
                "--log-dir",
                "-l",
                help="Path to the directory where logs will be stored. If the directory does not exist, it will be created.",
            ),
        ] = Path(f"migration_logs_{TODAY!s}"),
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Migrate Events to CogniteActivity."""
        client = EnvironmentVariables.create_from_environment().get_client()
        selected, dry_run, verbose = cls._prepare_asset_centric_arguments(
            client=client,
            mapping_file=mapping_file,
            data_set_id=data_set_id,
            consumption_view=consumption_view,
            ingestion_mapping=ingestion_mapping,
            dry_run=dry_run,
            verbose=verbose,
            kind="Events",
            resource_type="event",
            container_id=ContainerId("cdf_cdm", "CogniteActivity"),
        )

        cmd = MigrationCommand()

        cmd.run(
            lambda: cmd.migrate(
                selected=selected,
                data=AssetCentricMigrationIO(client),
                mapper=AssetCentricMapper(client),
                log_dir=log_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @classmethod
    def timeseries(
        cls,
        ctx: typer.Context,
        mapping_file: Annotated[
            Path | None,
            typer.Option(
                "--mapping-file",
                "-m",
                help="Path to the mapping file that contains the mapping from TimeSeries to CogniteTimeSeries. "
                "This file is expected to have the following columns: [id, dataSetId, space, externalId]."
                "The dataSetId is optional, and can be skipped. If it is set, it is used to check the access to the dataset.",
            ),
        ] = None,
        data_set_id: Annotated[
            str | None,
            typer.Option(
                "--data-set-id",
                "-s",
                help="The data set ID to select for the timeseries to migrate. If not provided and the mapping file is not provided"
                "an interactive selection will be performed to select the data set to migrate timeseries from.",
            ),
        ] = None,
        ingestion_mapping: Annotated[
            str | None,
            typer.Option(
                "--ingestion-mapping",
                "-i",
                help="The ingestion mapping to use for the migrated timeseries. If not provided, "
                "the default mapping to CogniteTimeSeries in CogniteCore will be used.",
            ),
        ] = None,
        consumption_view: Annotated[
            str | None,
            typer.Option(
                "--consumption-view",
                "-c",
                help="The consumption view to assign to the migrated timeseries Given as space:externalId/version. "
                "This will be used in Canvas to select which view to use when migrating timeseries. If not provided, "
                "CogniteTimeSeries in CogniteCore will be used.",
            ),
        ] = None,
        log_dir: Annotated[
            Path,
            typer.Option(
                "--log-dir",
                "-l",
                help="Path to the directory where logs will be stored. If the directory does not exist, it will be created.",
            ),
        ] = Path(f"migration_logs_{TODAY!s}"),
        skip_linking: Annotated[
            bool,
            typer.Option(
                "--skip-linking",
                "-x",
                help="If set, the migration will not create links between the old TimeSeries and the new CogniteTimeSeries.",
            ),
        ] = False,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Migrate TimeSeries to CogniteTimeSeries."""
        client = EnvironmentVariables.create_from_environment().get_client(enable_set_pending_ids=True)

        selected, dry_run, verbose = cls._prepare_asset_centric_arguments(
            client=client,
            mapping_file=mapping_file,
            data_set_id=data_set_id,
            consumption_view=consumption_view,
            ingestion_mapping=ingestion_mapping,
            dry_run=dry_run,
            verbose=verbose,
            kind="TimeSeries",
            resource_type="timeseries",
            container_id=ContainerId("cdf_cdm", "CogniteTimeSeries"),
        )
        if data_set_id is None and mapping_file is None:
            skip_linking = not questionary.confirm(
                "Do you want to link old and new TimeSeries?", default=not skip_linking
            ).ask()

        cmd = MigrationCommand()
        cmd.run(
            lambda: cmd.migrate(
                selected=selected,
                data=AssetCentricMigrationIO(client, skip_linking=skip_linking),
                mapper=AssetCentricMapper(client),
                log_dir=log_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @classmethod
    def files(
        cls,
        ctx: typer.Context,
        mapping_file: Annotated[
            Path | None,
            typer.Option(
                "--mapping-file",
                "-m",
                help="Path to the mapping file that contains the mapping from Files to CogniteFiles. "
                "This file is expected to have the following columns: [id, dataSetId, space, externalId]."
                "The dataSetId is optional, and can be skipped. If it is set, it is used to check the access to the dataset.",
            ),
        ] = None,
        data_set_id: Annotated[
            str | None,
            typer.Option(
                "--data-set-id",
                "-s",
                help="The data set ID to select for the files to migrate. If not provided, the dataSetId from the mapping file is used. "
                "If neither is provided, the default data set for the project is used.",
            ),
        ] = None,
        ingestion_mapping: Annotated[
            str | None,
            typer.Option(
                "--ingestion-mapping",
                "-i",
                help="The ingestion mapping to use for the migrated files. If not provided, "
                "the default mapping to CogniteFile in CogniteCore will be used.",
            ),
        ] = None,
        consumption_view: Annotated[
            str | None,
            typer.Option(
                "--consumption-view",
                "-c",
                help="The consumption view to assign to the migrated files Given as space:externalId/version. "
                "This will be used in Canvas to select which view to use when migrating timeseries. If not provided, "
                "CogniteFile in CogniteCore will be used.",
            ),
        ] = None,
        log_dir: Annotated[
            Path,
            typer.Option(
                "--log-dir",
                "-l",
                help="Path to the directory where logs will be stored. If the directory does not exist, it will be created.",
            ),
        ] = Path(f"migration_logs_{TODAY!s}"),
        skip_linking: Annotated[
            bool,
            typer.Option(
                "--skip-linking",
                "-x",
                help="If set, the migration will not create links between the old FileMetadata and the new CogniteFile.",
            ),
        ] = False,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Migrate Files to CogniteFiles."""
        client = EnvironmentVariables.create_from_environment().get_client(enable_set_pending_ids=True)

        selected, dry_run, verbose = cls._prepare_asset_centric_arguments(
            client=client,
            mapping_file=mapping_file,
            data_set_id=data_set_id,
            consumption_view=consumption_view,
            ingestion_mapping=ingestion_mapping,
            dry_run=dry_run,
            verbose=verbose,
            kind="FileMetadata",
            resource_type="file",
            container_id=ContainerId("cdf_cdm", "CogniteFile"),
        )
        cmd = MigrationCommand()

        if data_set_id is None:
            skip_linking = not questionary.confirm(
                "Do you want to link old and new Files?", default=not skip_linking
            ).ask()

        cmd.run(
            lambda: cmd.migrate(
                selected=selected,
                data=AssetCentricMigrationIO(client, skip_linking=skip_linking),
                mapper=AssetCentricMapper(client),
                log_dir=log_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @staticmethod
    def canvas(
        ctx: typer.Context,
        external_id: Annotated[
            list[str] | None,
            typer.Argument(
                help="The external ID of the Canvas to migrate. If not provided, and interactive selection will be "
                "performed to select the Canvas to migrate."
            ),
        ] = None,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of "
                "what would be done is printed. This is useful for checking that all resources referenced by the Canvas"
                "have been migrated to the new data modeling resources in CDF.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Migrate Canvas applications from Asset-Centric to data modeling in CDF.

        This command expects that the CogniteMigration data model is already deployed, and that the Mapping view
        is populated with the mapping from Asset-Centric resources to the new data modeling resources.
        """
        client = EnvironmentVariables.create_from_environment().get_client()

        cmd = MigrationCanvasCommand()
        cmd.run(
            lambda: cmd.migrate_canvas(
                client,
                external_ids=external_id,
                dry_run=dry_run,
                verbose=verbose,
            )
        )

    @staticmethod
    def infield_configs(
        ctx: typer.Context,
        output_dir: Annotated[
            Path,
            typer.Option(
                "--output-dir",
                "-o",
                help="Path to the directory where the Infield V2 configuration definitions will be dumped. It is recommended "
                "to govern these configurations in a git repository.",
            ),
        ] = Path("tmp"),
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-d",
                help="If set, the migration will not be executed, but only a report of what would be done is printed.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """Creates Infield V2 configurations from existing APM Configurations in CDF."""
        client = EnvironmentVariables.create_from_environment().get_client()

        cmd = MigrationCommand()
        cmd.run(
            lambda: cmd.create(
                client,
                creator=InfieldV2ConfigCreator(client),
                output_dir=output_dir,
                dry_run=dry_run,
                verbose=verbose,
            )
        )





import itertools
import re
import sys
from collections.abc import Iterator, MutableSequence
from functools import lru_cache
from pathlib import Path

from packaging.version import Version
from packaging.version import parse as parse_version
from rich import print

from cognite_toolkit._cdf_tk.builders import get_loader
from cognite_toolkit._cdf_tk.constants import DOCKER_IMAGE_NAME
from cognite_toolkit._cdf_tk.data_classes import ModuleDirectories
from cognite_toolkit._cdf_tk.utils import iterate_modules, read_yaml_file, safe_read, safe_write
from cognite_toolkit._version import __version__

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

class Change:
    """A change is a single migration step that can be applied to a project."""

    deprecated_from: Version
    required_from: Version | None = None
    has_file_changes: bool = False

    def __init__(self, organization_dir: Path, workflow_dir: Path | None) -> None:
        self._organization_dir = organization_dir
        self._workflow_dir = workflow_dir


class AutomaticChange(Change):
    """An automatic change is a change that can be applied automatically to a project."""

    def do(self) -> set[Path]:
        return set()


class ManualChange(Change):
    """A manual change is a change that requires manual intervention to be applied to a project."""

    def needs_to_change(self) -> set[Path]:
        return set()

    def instructions(self, files: set[Path]) -> str:
        return ""


class SetKindOnFile(AutomaticChange):
    """Adds the kind to the filename of all resource files.

Before `your_file.yaml`:
After `your_file.FileMetadata.yaml`:
    """

    deprecated_from = Version("0.4.0")
    has_file_changes = True

    def do(self) -> set[Path]:
        module_directories = ModuleDirectories.load(self._organization_dir)
        changed: set[Path] = set()
        for module in module_directories:
            for resource_folder, source_files in module.source_paths_by_resource_folder.items():
                for source_file in source_files:
                    if source_file.suffix not in {".yaml", ".yml"}:
                        continue
                    loader, _ = get_loader(source_file, resource_folder, force_pattern=True)
                    if loader is None:
                        print(f"Could not find loader for {source_file}")
                        continue
                    if source_file.stem.casefold().endswith(loader.kind.casefold()):
                        continue
                    new_name = source_file.with_name(f"{source_file.stem}.{loader.kind}{source_file.suffix}")
                    source_file.rename(new_name)
                    changed.add(source_file)
                    for suffix in [".sql", ".csv", ".parquet"]:
                        if (adjacent_file := source_file.with_suffix(suffix)).exists():
                            adjacent_file.rename(new_name.with_suffix(suffix))
                            changed.add(adjacent_file)

        return changed


class FixViewBasedLocationFilter(AutomaticChange):
    """The created view-based location filter has been fixed to be compatible with the CDF API.

Before `your.LocationFilter.yaml`:
```yaml
...
views:
  externalId: my_view
  space: my_space
  version: "1"
  representsEntity: ASSET
...
```
After `your.LocationFilter.yaml`:
```yaml
...
views:
  - externalId: my_view
    space: my_space
    version: "1"
    representsEntity: ASSET
...
    """

    deprecated_from = Version("0.3.1")
    required_from = Version("0.3.1")
    has_file_changes = True

    def do(self) -> set[Path]:
        changed = set()
        for resource_yaml in itertools.chain(self._organization_dir.rglob("*LocationFilter.yaml"), self._organization_dir.rglob("*LocationFilter.yml")):
            content = safe_read(resource_yaml)
            if "views:" not in content:
                continue
            lines = content.splitlines()
            new_lines: list[str] = []
            is_next = False
            indent: int | None = None
            for line in lines:
                if line.startswith("views:"):
                    is_next = True
                elif is_next:
                    indent = len(line) - len(line.lstrip())
                    line = f"{indent * ' '}- {line[indent:]}"
                    is_next = False
                elif indent is not None:
                    line_indent = len(line) - len(line.lstrip())
                    if line_indent == indent:
                        line = f"  {line}"
                    else:
                        indent = None
                new_lines.append(line)
            safe_write(resource_yaml, "\n".join(new_lines))
            changed.add(resource_yaml)
        return changed



class NodeAPICallParametersNoLongerSupported(AutomaticChange):
    """Setting API call parameters in the 'node' section of the config files is no longer supported.

This is now handled correctly by the CDF Toolkit and should be removed from the config files.

For example, in data_models/my_node.node.yaml, before:
```yaml
replace: true
nodes:
 - space: node_space
   externalId: default_infield_config_minimal
```
After:
```yaml
- space: node_space
  externalId: default_infield_config_minimal
```
    """

    deprecated_from = Version("0.3.0b1")
    required_from = Version("0.3.0b1")
    has_file_changes = True

    def do(self) -> set[Path]:
        from cognite_toolkit._cdf_tk.utils import resource_folder_from_path

        api_call_parameters = {
            "skipOnVersionConflict", "replace", "autoCreateDirectRelations"
        }

        changed: set[Path] = set()
        resource_yaml: Path
        for resource_yaml in self._organization_dir.rglob("*.yaml"):
            try:
                resource_folder = resource_folder_from_path(resource_yaml)
            except ValueError:
                continue
            if resource_folder == "data_models" and resource_yaml.stem.casefold().endswith("node"):
                content = safe_read(resource_yaml)
                has_api_call_parameters = False
                new_content:  list[str] = []
                indent: int | None = None
                for line in content.splitlines():
                    if any(line.startswith(parameter) for parameter in api_call_parameters):
                        has_api_call_parameters = True
                        continue
                    if (line.startswith("nodes:") or line.startswith("node:")) and has_api_call_parameters:
                        continue
                    if has_api_call_parameters and indent is None:
                        indent = len(line) - len(line.lstrip())
                    if has_api_call_parameters and indent is not None:
                        line = line[indent:]
                    new_content.append(line)
                if has_api_call_parameters:
                    changed.add(resource_yaml)
                    safe_write(resource_yaml, "\n".join(new_content))

        return changed


class ResourceFolderTimeSeriesDatapointsRemoved(AutomaticChange):
    """The resource folder 'timeseries_datapoints' have been removed.

The `csv` and `parquet` files that were previously stored in the `timeseries_datapoints` folder
should be moved to the `timeseries` folder.

Before:
```bash
    my_module/
       timeseries_datapoints/
          my_datapoints.Datapoints.csv
```
After:
```bash
    my_module/
       timeseries/
          my_datapoints.Datapoints.csv
```
    """

    deprecated_from = Version("0.3.0a7")
    required_from = Version("0.3.0a7")
    has_file_changes = True

    def do(self) -> set[Path]:
        changed = set()
        for module, source_files in iterate_modules(self._organization_dir):
            for resource_dir in module.iterdir():
                if resource_dir.name == "timeseries_datapoints":
                    (module / "timeseries").mkdir(exist_ok=True)
                    for filepath in resource_dir.rglob("*"):
                        target = module / "timeseries" / filepath.relative_to(resource_dir)
                        target.parent.mkdir(exist_ok=True, parents=True)
                        if target.exists():
                            print(f'  [bold red]ERROR ([/][red] Cannot move file [/][bold red]{filepath}[/][red] to [/][bold red]{target}[/][red]): File already exists')
                            continue
                        filepath.rename(target)
                        changed.add(target)
        return changed


class AuthVerifySplit(AutomaticChange):
    """The `cdf auth verify` has been split into `cdf auth init` and `cdf auth verify`.

The `cdf auth init` command initializes the authorization for a user/service principal to run the CDF Toolkit commands,
it will by default also verify the capabilities after the initialization. Thus it replaces the `cdf auth verify` command.
In addition, the `cdf auth verify` command will only verify the capabilities without initializing the authorization.
    """

    deprecated_from = Version("0.3.0a4")
    required_from = Version("0.3.0a4")
    has_file_changes = False


class DeployCleanInteractiveFlagRemoved(AutomaticChange):
    """The `--interactive` flag has been removed from the `cdf deploy` and `cdf clean` commands.
    """

    deprecated_from = Version("0.3.0a4")
    required_from = Version("0.3.0a4")
    has_file_changes = False


class SharedVerboseFlagRemoved(AutomaticChange):
    """The shared `--verbose` flag been removed. Now each command has its own `--verbose` flag.

For example, before:
```bash
    cdf --verbose deploy
```

After:
```bash
    cdf deploy --verbose
```
    """

    deprecated_from = Version("0.3.0a4")
    required_from = Version("0.3.0a4")
    has_file_changes = False


class RenamedOrganizationDirInCDFToml(AutomaticChange):
    """In the cdf.toml file, the 'organization_dir' field in the 'cdf' section has been renamed to
'default_organization_dir'.

In cdf.toml, before:
```toml
[cdf]
organization_dir = "my_organization"
```
After:
```toml
[cdf]
default_organization_dir = "my_organization"
```
"""

    deprecated_from = Version("0.3.0a3")
    required_from = Version("0.3.0a3")
    has_file_changes = True

    def do(self) -> set[Path]:
        cdf_toml = Path.cwd() / "cdf.toml"
        if not cdf_toml.exists():
            return set()
        raw = safe_read(cdf_toml)
        new_cdf_toml = []
        changes: set[Path] = set()
        # We do not parse the TOML file to avoid removing comments
        is_after_cdf_section=False
        for line in raw.splitlines():
            if line.startswith("[cdf]"):
                is_after_cdf_section = True
            if line.startswith("organization_dir = ") and is_after_cdf_section:
                new_line = line.replace("organization_dir", "default_organization_dir")
                new_cdf_toml.append(new_line)
                if new_line != line:
                    changes.add(cdf_toml)
            else:
                new_cdf_toml.append(line)
        cdf_toml.write_text("\n".join(new_cdf_toml), encoding="utf-8")
        return changes


class InitCommandReplaced(AutomaticChange):
    """The `cdf-tk init` has been replaced by `cdf repo init` and `cdf modules init`.

The `cdf repo init` command initializer the current directory with config and git files such as .gitignore.
The `cdf modules init` has an interactive prompt for the user to select the modules to include.
    """

    deprecated_from = Version("0.3.0a1")
    required_from = Version("0.3.0a1")
    has_file_changes = False


class SystemYAMLReplaced(AutomaticChange):
    """The _system.yaml file is now replaced by cdf.toml in the cwd of the project.

Before:
```bash
    my_organization/
        _system.yaml
```
After:
```bash
    cdf.toml
    my_organization/
```
    """

    deprecated_from = Version("0.3.0a1")
    required_from = Version("0.3.0a1")
    has_file_changes = True

    def do(self) -> set[Path]:
        # Avoid circular import
        from .modules import ModulesCommand

        system_yaml = self._organization_dir / "_system.yaml"
        if not system_yaml.exists():
            system_yaml = self._organization_dir / "cognite_modules" / "_system.yaml"
            if not system_yaml.exists():
                return set()
        content = read_yaml_file(system_yaml)
        current_version = content.get("cdf_toolkit_version", __version__)

        cdf_toml_content = ModulesCommand(skip_tracking=True).create_cdf_toml(self._organization_dir)
        cdf_toml_content = cdf_toml_content.replace(f'version = "{__version__}"', f'version = "{current_version}"')

        cdf_toml_path = Path.cwd() / "cdf.toml"
        cdf_toml_path.write_text(cdf_toml_content, encoding="utf-8")
        system_yaml.unlink()
        return {cdf_toml_path, system_yaml}


class ResourceFolderLabelsRenamed(AutomaticChange):
    """The resource folder 'labels' have been renamed to 'classic'.

Before:
```bash
    my_module/
       labels/
          my_labels.Label.yaml
```
After:
```bash
    my_module/
       classic/
          my_labels.Label.yaml
```
    """

    deprecated_from = Version("0.3.0a1")
    required_from = Version("0.3.0a1")
    has_file_changes = True

    def do(self) -> set[Path]:
        changed = set()
        for module, source_files in iterate_modules(self._organization_dir):
            for resource_dir in module.iterdir():
                if resource_dir.name == "labels":
                    (module / "classic").mkdir(exist_ok=True)
                    for files in resource_dir.rglob("*"):
                        target = module / "classic" / files.relative_to(resource_dir)
                        target.parent.mkdir(exist_ok=True, parents=True)
                        files.rename(target)
                        changed.add(target)
        return changed


class RenamedModulesSection(AutomaticChange):
    """The 'modules' section in the config files has been renamed to 'variables'.
This change updates the config files to use the new name.

For example in config.dev.yaml, before:
```yaml
    modules:
      cognite_modules:
        cdf_cluster: ${CDF_CLUSTER}
        cicd_clientId: ${IDP_CLIENT_ID}
        cicd_clientSecret: ${IDP_CLIENT_SECRET}
```
After:
```yaml
    variables:
      cognite_modules:
        cdf_cluster: ${CDF_CLUSTER}
        cicd_clientId: ${IDP_CLIENT_ID}
        cicd_clientSecret: ${IDP_CLIENT_SECRET}
```
    """

    deprecated_from = Version("0.2.0a3")
    required_from = Version("0.2.0a3")
    has_file_changes = True

    def do(self) -> set[Path]:
        changed: set[Path] = set()
        for config_yaml in self._organization_dir.glob("config.*.yaml"):
            data_raw = safe_read(config_yaml)
            # We do not parse the YAML file to avoid removing comments
            updated_file: list[str] = []
            for line in data_raw.splitlines():
                if line.startswith("modules:"):
                    changed.add(config_yaml)
                    updated_file.append(line.replace("modules:", "variables:"))
                else:
                    updated_file.append(line)
            safe_write(config_yaml, "\n".join(updated_file))
        return changed


class BuildCleanFlag(AutomaticChange):
    """The `cdf-tk build` command no longer accepts the `--clean` flag.

The build command now always cleans the build directory before building.
To avoid cleaning the build directory, you can use the `--no-clean` flag.
    """

    deprecated_from = Version("0.2.0a3")
    required_from = Version("0.2.0a3")
    has_file_changes = False


class CommonFunctionCodeNotSupported(ManualChange):
    """."""

    deprecated_from = Version("0.2.0a4")
    required_from = Version("0.2.0a4")
    has_file_changes = True

    def needs_to_change(self) -> set[Path]:
        common_function_code = self._organization_dir / "common_function_code"
        if not common_function_code.exists():
            return set()
        needs_change = {common_function_code}
        for py_file in self._organization_dir.rglob("*.py"):
            content = safe_read(py_file).splitlines()
            use_common_function_code = any(
                (line.startswith("from common") or line.startswith("import common")) for line in content
            )
            if use_common_function_code:
                needs_change.add(py_file)
        return needs_change

    def instructions(self, files: set[Path]) -> str:
        to_update = []
        for module, py_files in itertools.groupby(sorted(files, key=self.get_module_name), key=self.get_module_name):
            if module == Path("."):
                # This is the common_function_code folder
                continue
            to_update.append(f"  - In module {module.relative_to(self._organization_dir).as_posix()!r}:")
            for py_file in py_files:
                to_update.append(f"    - In file {py_file.relative_to(module).as_posix()!r}")
        to_update_str = "\n".join(to_update)
        return (
            "Cognite-Toolkit no longer supports the common functions code.\n"
            f"Please update the following files to not use 'common' module:\n{to_update_str}"
            f"\n\nThen remove the '{self._organization_dir.name}/common_function_code' folder."
        )

    @staticmethod
    @lru_cache(maxsize=128)
    def get_module_name(file_path: Path) -> Path:
        while file_path.parent != file_path:
            if file_path.name == "functions":
                return file_path.parent
            file_path = file_path.parent
        return Path(".")


class FunctionExternalDataSetIdRenamed(AutomaticChange):
    """The 'externalDataSetId' field in function YAML files has been renamed to 'dataSetExternalId'.
This change updates the function YAML files to use the new name.

The motivation for this change is to make the naming consistent with the rest of the Toolkit.

For example, in functions/my_function.yaml, before:
```yaml
externalDataSetId: my_external_id
```
After:
```yaml
dataSetExternalId: my_external_id
```
    """

    deprecated_from = Version("0.2.0a5")
    required_from = Version("0.2.0a5")
    has_file_changes = True

    def do(self) -> set[Path]:
        changed: set[Path] = set()
        for resource_yaml in self._organization_dir.glob("*.yaml"):
            if resource_yaml.parent == "functions":
                content = safe_read(resource_yaml)
                if "externalDataSetId" in content:
                    changed.add(resource_yaml)
                    content = content.replace("externalDataSetId", "dataSetExternalId")
                    safe_write(resource_yaml, content)
        return changed


class ConfigYAMLSelectedRenaming(AutomaticChange):
    """The 'environment.selected_modules_and_packages' field in the config.yaml files has been
renamed to 'selected'.
This change updates the config files to use the new name.

For example, in config.dev.yaml, before:
```yaml
environment:
  selected_modules_and_packages:
    - my_module
```
After:
```yaml
environment:
  selected:
    - my_module
```
    """

    deprecated_from = Version("0.2.0b1")
    has_file_changes = True

    def do(self) -> set[Path]:
        changed = set()
        for config_yaml in self._organization_dir.glob("config.*.yaml"):
            data = safe_read(config_yaml)
            if "selected_modules_and_packages" in data:
                changed.add(config_yaml)
                data = data.replace("selected_modules_and_packages", "selected")
                safe_write(config_yaml, data)
        return changed


class RequiredFunctionLocation(AutomaticChange):
    """Function Resource YAML files are now expected to be in the 'functions' folder.
Before they could be in subfolders inside the 'functions' folder.

This change moves the function YAML files to the 'functions' folder.

For example, before:
```bash
    modules/
      my_module/
          functions/
            some_subdirectory/
                my_function.yaml
```
After:
```bash
    modules/
      my_module/
          functions/
            my_function.yaml
```
    """

    deprecated_from = Version("0.2.0b3")
    required_from = Version("0.2.0b3")
    has_file_changes = True

    def do(self) -> set[Path]:
        changed = set()
        for resource_yaml in self._organization_dir.glob("functions/**/*.yaml"):
            if self._is_function(resource_yaml):
                new_path = self._new_path(resource_yaml)
                if new_path != resource_yaml:
                    resource_yaml.rename(new_path)
                    changed.add(new_path)
        return changed

    @staticmethod
    def _is_function(resource_yaml: Path) -> bool:
        # Functions require a 'name' field and to distinguish from a FunctionSchedule
        # we check that the 'cronExpression' field is not present
        parsed = read_yaml_file(resource_yaml)
        if isinstance(parsed, dict):
            return "name" in parsed and "cronExpression" not in parsed
        elif isinstance(parsed, list):
            return all("name" in item and "cronExpression" not in item for item in parsed)
        return False

    @staticmethod
    def _new_path(resource_yaml: Path) -> Path:
        # Search the path for the 'functions' folder and move the file there
        for parent in resource_yaml.parents:
            if parent.name == "functions":
                return parent / resource_yaml.name
        return resource_yaml


class UpdateModuleVersion(AutomaticChange):
    deprecated_from = parse_version(__version__)
    required_from = parse_version(__version__)
    has_file_changes = True

    def do(self) -> set[Path]:
        cdf_toml = Path.cwd() / "cdf.toml"
        if not cdf_toml.exists():
            return set()
        raw = safe_read(cdf_toml)
        new_cdf_toml = []
        changes: set[Path] = set()
        # We do not parse the TOML file to avoid removing comments
        is_after_module_section=False
        for line in raw.splitlines():
            if line.startswith("[modules]"):
                is_after_module_section = True
            if line.startswith("version = ") and is_after_module_section:
                new_line = f'version = "{__version__}"'
                new_cdf_toml.append(new_line)
                if new_line != line:
                    changes.add(cdf_toml)
            else:
                new_cdf_toml.append(line)
        safe_write(cdf_toml, "\n".join(new_cdf_toml))
        return changes

class UpdateDockerImageVersion(AutomaticChange):
    deprecated_from = parse_version(__version__)
    required_from = parse_version(__version__)
    has_file_changes = True

    def do(self) -> set[Path]:
        if self._workflow_dir is None:
            return set()
        changed = set()
        for workflow_file in itertools.chain(self._workflow_dir.rglob("*.yaml"), self._workflow_dir.rglob("*.yml")):
            content = safe_read(workflow_file)
            new_content = re.sub(rf"image: {DOCKER_IMAGE_NAME}:[0-9.ab]+", f"image: {DOCKER_IMAGE_NAME}:{__version__}", content)
            if new_content != content:
                safe_write(workflow_file, new_content)
                changed.add(workflow_file)
        return changed

UPDATE_MODULE_VERSION_DOCSTRING = """In the cdf.toml file, the 'version' field in the 'module' section has been updated to the same version as the CLI.

This change updated the 'version' field in the cdf.toml file to the same version as the CLI.

In cdf.toml, before:
```toml
[modules]
version = "{module_version}"
```
After:
```toml
[modules]
version = "{cli_version}"
```
    """
UpdateModuleVersion.__doc__ = UPDATE_MODULE_VERSION_DOCSTRING

UPDATE_IMAGE_VERSION_DOCSTRING = """Update the docker image version in the workflow files to the current version of the CLI.

Before:
```yaml
image: cognite/toolkit:{module_version}
```
After:
```yaml
image: cognite/toolkit:{cli_version}
```
    """
UpdateDockerImageVersion.__doc__ = UPDATE_IMAGE_VERSION_DOCSTRING


_CHANGES: list[type[Change]] = [
    change for change in itertools.chain(AutomaticChange.__subclasses__(), ManualChange.__subclasses__())
]


class Changes(list, MutableSequence[Change]):
    @classmethod
    def load(cls, module_version: Version, project_path: Path, workflow_dir: Path | None) -> Self:
        return cls([change(project_path, workflow_dir) for change in _CHANGES if change.deprecated_from >= module_version])

    @property
    def required_changes(self) -> "Changes":
        return Changes([change for change in self if change.required_from is not None])

    @property
    def optional_changes(self) -> "Changes":
        return Changes([change for change in self if change.required_from is None])

    def __iter__(self) -> Iterator[Change]:
        return super().__iter__()



import sys
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, ClassVar, TypeVar

from cognite_toolkit import _version
from cognite_toolkit._cdf_tk.constants import BUILD_ENVIRONMENT_FILE
from cognite_toolkit._cdf_tk.exceptions import ToolkitFileNotFoundError, ToolkitRequiredValueError, ToolkitVersionError
from cognite_toolkit._cdf_tk.utils import read_yaml_file

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@dataclass
class ConfigCore(ABC):
    """Base class for config files."""

    filename: ClassVar[str]
    filepath: Path

    @classmethod
    def get_filename(cls, build_env: str) -> str:
        return cls.filename.format(build_env=build_env)

    @classmethod
    def load_from_directory(cls, organization_dir: Path, build_env: str) -> Self:
        filename = cls.get_filename(build_env)
        filepath = organization_dir / filename
        filepath = filepath if filepath.is_file() else Path.cwd() / filename
        if not filepath.is_file():
            raise ToolkitFileNotFoundError(f"{filename!r} does not exist.")

        try:
            return cls.load(read_yaml_file(filepath), build_env, filepath)
        except KeyError as e:
            raise ToolkitRequiredValueError(f"Required field {e.args} is missing in {filename!r}.") from e

    @classmethod
    @abstractmethod
    def load(cls, data: dict[str, Any], build_env: str, filepath: Path) -> Self:
        raise NotImplementedError


T_BuildConfig = TypeVar("T_BuildConfig", bound=ConfigCore)


def _load_version_variable(data: dict[str, Any], file_name: str) -> str:
    try:
        cdf_tk_version: str = data["cdf_toolkit_version"]
    except KeyError:
        err_msg = f"System variables are missing required field 'cdf_toolkit_version' in {file_name!s}. {{}}"
        if file_name == BUILD_ENVIRONMENT_FILE:
            raise ToolkitVersionError(
                err_msg.format("Rerun `cdf build` to build the modules again and create it correctly.")
            )
        raise ToolkitVersionError(
            err_msg.format("Run `cdf modules upgrade` to initialize the modules again to create a correct file.")
        )

    if cdf_tk_version != _version.__version__:
        raise ToolkitVersionError(
            f"The version of the modules ({cdf_tk_version}) does not match the version of the installed CLI "
            f"({_version.__version__}). Please either run `cdf modules upgrade` to upgrade the modules OR "
            f"run `pip install cognite-toolkit=={cdf_tk_version}` to downgrade cdf-tk CLI."
        )
    return cdf_tk_version



import re
from typing import Any

from pydantic import Field, field_validator, model_serializer
from pydantic_core.core_schema import SerializationInfo, SerializerFunctionWrapHandler

from cognite_toolkit._cdf_tk.constants import (
    CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN,
    DM_EXTERNAL_ID_PATTERN,
    DM_VERSION_PATTERN,
    FORBIDDEN_CONTAINER_AND_VIEW_EXTERNAL_IDS,
    FORBIDDEN_CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER,
    SPACE_FORMAT_PATTERN,
)
from cognite_toolkit._cdf_tk.utils.collection import humanize_collection

from .base import ToolkitResource
from .view_field_definitions import ViewProperty, ViewReference

KEY_PATTERN = re.compile(CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN)


class ViewYAML(ToolkitResource):
    space: str = Field(
        description="Id of the space that the view belongs to.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the view.",
        min_length=1,
        max_length=255,
        pattern=DM_EXTERNAL_ID_PATTERN,
    )
    version: str = Field(
        description="Version of the view.",
        max_length=43,
        pattern=DM_VERSION_PATTERN,
    )
    name: str | None = Field(
        default=None,
        description="name for the view.",
        max_length=255,
    )
    description: str | None = Field(
        default=None,
        description="Description of the view.",
        max_length=1024,
    )
    filter: dict[str, Any] | None = Field(
        default=None,
        description="A filter Domain Specific Language (DSL) used to create advanced filter queries.",
    )
    implements: list[ViewReference] | None = Field(
        default=None,
        description="References to the views from where this view will inherit properties.",
    )
    properties: dict[str, ViewProperty] | None = Field(
        default=None, description="Set of properties to apply to the View."
    )

    @field_validator("external_id")
    @classmethod
    def check_forbidden_external_id_value(cls, val: str) -> str:
        """Check the external_id not present in forbidden set"""
        if val in FORBIDDEN_CONTAINER_AND_VIEW_EXTERNAL_IDS:
            raise ValueError(
                f"'{val}' is a reserved view External ID. Reserved External IDs are: {humanize_collection(FORBIDDEN_CONTAINER_AND_VIEW_EXTERNAL_IDS)}"
            )
        return val

    @field_validator("properties")
    @classmethod
    def validate_properties_identifier(cls, val: dict[str, str]) -> dict[str, str]:
        """Validate properties Identifier"""
        for key in val.keys():
            if not KEY_PATTERN.match(key):
                raise ValueError(f"Property '{key}' does not match the required pattern: {KEY_PATTERN.pattern}")
            if key in FORBIDDEN_CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER:
                raise ValueError(
                    f"'{key}' is a reserved property identifier. Reserved identifiers are: {humanize_collection(FORBIDDEN_CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER)}"
                )
        return val

    @model_serializer(mode="wrap")
    def serialize_container(self, handler: SerializerFunctionWrapHandler, info: SerializationInfo) -> dict:
        serialized_data = handler(self)
        if self.properties:
            serialized_data["properties"] = {
                key: value.model_dump(**vars(info)) for key, value in self.properties.items()
            }
        return serialized_data



from enum import Enum
from pathlib import Path
from typing import Annotated, Any, cast

import questionary
import typer
from rich import print

from cognite_toolkit._cdf_tk.commands import PurgeCommand
from cognite_toolkit._cdf_tk.exceptions import ToolkitValueError
from cognite_toolkit._cdf_tk.feature_flags import Flags
from cognite_toolkit._cdf_tk.storageio.selectors import (
    InstanceFileSelector,
    InstanceSelector,
    InstanceViewSelector,
    SelectedView,
)
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from cognite_toolkit._cdf_tk.utils.cli_args import parse_view_str
from cognite_toolkit._cdf_tk.utils.interactive_select import AssetInteractiveSelect, DataModelingSelect


class InstanceTypeEnum(str, Enum):
    node = "node"
    edge = "edge"


class PurgeApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.command("dataset")(self.purge_dataset)
        self.command("space")(self.purge_space)
        if Flags.PURGE_INSTANCES.is_enabled() or Flags.v07.is_enabled():
            self.command("instances")(self.purge_instances)

    def main(self, ctx: typer.Context) -> None:
        """Commands deleting data from Cognite Data Fusion."""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf purge --help[/] for more information.")

    @staticmethod
    def purge_dataset(
        ctx: typer.Context,
        external_id: Annotated[
            str | None,
            typer.Argument(
                help="External id of the dataset to purge. If not provided, interactive mode will be used.",
            ),
        ] = None,
        include_dataset: Annotated[
            bool,
            typer.Option(
                "--include-dataset",
                "-i",
                help="Whether to archive the dataset itself after purging its contents.",
                hidden=Flags.v07.is_enabled(),
            ),
        ] = False,
        archive_dataset: Annotated[
            bool,
            typer.Option(
                "--archive-dataset",
                help="Whether to archive the dataset itself after purging its contents.",
                hidden=not Flags.v07.is_enabled(),
            ),
        ] = False,
        skip_data: Annotated[
            bool,
            typer.Option(
                "--skip-data",
                "-s",
                help="Skip deleting the data in the dataset, only delete configurations. The resources that are "
                "considered data are: time series, event, files, assets, sequences, relationships, "
                "labels, and 3D Models",
                hidden=not Flags.v07.is_enabled(),
            ),
        ] = False,
        include_configurations: Annotated[
            bool,
            typer.Option(
                "--include-configurations",
                "-c",
                help="Include configurations, workflows, extraction pipelines and transformations in the purge.",
                hidden=not Flags.v07.is_enabled(),
            ),
        ] = False,
        asset_recursive: Annotated[
            bool,
            typer.Option(
                "--asset-recursive",
                "-a",
                help="When deleting assets, delete all child assets recursively. CAVEAT: This can lead to assets"
                " not in the selected dataset being deleted if they are children of assets in the dataset.",
                hidden=not Flags.v07.is_enabled(),
            ),
        ] = False,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-r",
                help="Whether to do a dry-run, do dry-run if present.",
            ),
        ] = False,
        auto_yes: Annotated[
            bool,
            typer.Option(
                "--yes",
                "-y",
                help="Automatically confirm that you are sure you want to purge the dataset.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will delete the contents of the specified dataset"""
        cmd = PurgeCommand()
        client = EnvironmentVariables.create_from_environment().get_client()

        if external_id is None:
            # Is Interactive
            interactive = AssetInteractiveSelect(client, operation="purge")
            external_id = interactive.select_data_set(allow_empty=False)
            if Flags.v07.is_enabled():
                skip_data = not questionary.confirm(
                    "Delete data in the dataset (time series, events, files, assets, sequences, relationships, labels, 3D models)?",
                    default=True,
                ).ask()
                include_configurations = questionary.confirm(
                    "Delete configurations (workflows, extraction pipelines and transformations) in the dataset?",
                    default=False,
                ).ask()
                asset_recursive = questionary.confirm(
                    "When deleting assets, delete all child assets recursively? (WARNING: This can lead "
                    "to assets not in the selected dataset being deleted if they are children of assets in the dataset.)",
                    default=False,
                ).ask()
            archive_dataset = questionary.confirm("Archive the dataset itself after purging?", default=False).ask()
            dry_run = questionary.confirm("Dry run?", default=True).ask()
            verbose = questionary.confirm("Verbose?", default=True).ask()

            user_options = [archive_dataset, dry_run, verbose]
            if Flags.v07.is_enabled():
                user_options.extend([skip_data, include_configurations, asset_recursive])

            if any(selected is None for selected in user_options):
                raise typer.Abort("Aborted by user.")

        else:
            archive_dataset = archive_dataset if Flags.v07.is_enabled() else include_dataset

        if not Flags.v07.is_enabled():
            skip_data = False
            include_configurations = True
            asset_recursive = False

        cmd.run(
            lambda: cmd.dataset(
                client,
                external_id,
                archive_dataset,
                not skip_data,
                include_configurations,
                asset_recursive,
                dry_run,
                auto_yes,
                verbose,
            )
        )

    @staticmethod
    def purge_space(
        ctx: typer.Context,
        space: Annotated[
            str | None,
            typer.Argument(
                help="Space to purge. If not provided, interactive mode will be used.",
            ),
        ] = None,
        include_space: Annotated[
            bool,
            typer.Option(
                "--include-space",
                "-i",
                help="Include space in the purge. This will also delete the space.",
            ),
        ] = False,
        delete_datapoints: Annotated[
            bool,
            typer.Option(
                "--delete-datapoints",
                help="Delete datapoints linked to CogniteTimeSeries nodes in the space.",
                hidden=not Flags.v07.is_enabled(),
            ),
        ] = False,
        delete_file_content: Annotated[
            bool,
            typer.Option(
                "--delete-file-content",
                help="Delete file content linked to CogniteFile nodes in the space.",
                hidden=not Flags.v07.is_enabled(),
            ),
        ] = False,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-r",
                help="Whether to do a dry-run, do dry-run if present.",
            ),
        ] = False,
        auto_yes: Annotated[
            bool,
            typer.Option(
                "--yes",
                "-y",
                help="Automatically confirm that you are sure you want to purge the space.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will delete the contents of the specified space."""

        cmd = PurgeCommand()
        client = EnvironmentVariables.create_from_environment().get_client()

        if space is None:
            # Is Interactive
            interactive = DataModelingSelect(client, operation="purge")
            space_type = interactive.select_space_type()
            if space_type == "empty":
                space = interactive.select_empty_spaces(multiselect=False)
            elif space_type == "instance":
                space = interactive.select_instance_space(multiselect=False)
            elif space_type == "schema":
                space = interactive.select_schema_space(include_global=False).space
            else:
                raise ToolkitValueError("Invalid space type selected.")
            dry_run = questionary.confirm("Dry run?", default=True).ask()
            if space_type == "empty":
                include_space = True
            else:
                include_space = questionary.confirm("Delete the space itself?", default=False).ask()

        cmd.run(
            lambda: cmd.space(
                client=client,
                selected_space=space,
                include_space=include_space,
                delete_datapoints=delete_datapoints,
                delete_file_content=delete_file_content,
                dry_run=dry_run,
                auto_yes=auto_yes,
                verbose=verbose,
            )
        )

    @staticmethod
    def purge_instances(
        view: Annotated[
            str | None,
            typer.Argument(
                help="Purge instances with properties in the specified view. Expected format is "
                "'space:externalId/version'. For example 'cdf_cdm:CogniteTimeSeries/v1' will purge all nodes"
                "that have properties in the CogniteTimeSeries view. If not provided and no "
                "instance list is provided, interactive mode will be used.",
            ),
        ] = None,
        instance_space: Annotated[
            list[str] | None,
            typer.Option(
                "--instance-space",
                "-s",
                help="Only purge instances that are in the specified instance space(s).",
            ),
        ] = None,
        instance_type: Annotated[
            InstanceTypeEnum,
            typer.Option(
                "--instance-type",
                "-t",
                help="Type of instances to purge. Can be 'node' or 'edge'. Default is 'node'.",
                case_sensitive=False,
                show_default=True,
            ),
        ] = InstanceTypeEnum.node,
        instance_list: Annotated[
            Path | None,
            typer.Option(
                "--list",
                "-l",
                help="Path to a file containing a list of instance external IDs to purge. This file should be a csv file with space,externalId,instanceType columns."
                "If this option is provided, the view, instance-space and instance-type options will be ignored.",
                exists=True,
                file_okay=True,
                dir_okay=False,
                readable=True,
                resolve_path=True,
                show_default=True,
            ),
        ] = None,
        dry_run: Annotated[
            bool,
            typer.Option(
                "--dry-run",
                "-r",
                help="Whether to do a dry-run, do dry-run if present.",
            ),
        ] = False,
        unlink: Annotated[
            bool,
            typer.Option(
                "--skip-unlink",
                "-u",
                help="This only applies to CogniteTimeSeries and CogniteFile nodes. By default, the purge command will unlink the "
                "node from the datapoints/file content before deleting the node. If you want to delete the nodes with their datapoints/file content, "
                "you can skip the unlinking. Note that this will delete the datapoints/file content "
                "themselves, not the links to their parent nodes.",
            ),
        ] = True,
        auto_yes: Annotated[
            bool,
            typer.Option(
                "--yes",
                "-y",
                help="Automatically confirm that you are sure you want to purge the instances.",
            ),
        ] = False,
        verbose: Annotated[
            bool,
            typer.Option(
                "--verbose",
                "-v",
                help="Turn on to get more verbose output when running the command",
            ),
        ] = False,
    ) -> None:
        """This command will delete the contents of the specified instances."""

        cmd = PurgeCommand()
        client = EnvironmentVariables.create_from_environment().get_client(enable_set_pending_ids=True)
        is_interactive = view is None and instance_list is None
        selector: InstanceSelector
        if is_interactive:
            interactive = DataModelingSelect(client, operation="purge")
            select_view = interactive.select_view(include_global=True)
            selected_instance_type = interactive.select_instance_type(select_view.used_for)
            instance_space = interactive.select_instance_space(True, select_view.as_id(), selected_instance_type)
            selector = InstanceViewSelector(
                view=SelectedView(
                    space=select_view.space, external_id=select_view.external_id, version=select_view.version
                ),
                instance_type=selected_instance_type,
                instance_spaces=tuple(instance_space) if instance_space else None,
            )
            dry_run = questionary.confirm("Dry run?", default=True).ask()
            unlink = questionary.confirm("Unlink instances connected to timeseries or files?", default=True).ask()
        elif instance_list is not None:
            selector = InstanceFileSelector(datafile=instance_list)
        elif view is not None:
            view_id = parse_view_str(view)
            selector = InstanceViewSelector(
                view=SelectedView(
                    space=view_id.space, external_id=view_id.external_id, version=cast(str, view_id.version)
                ),
                instance_type=instance_type.value,
                instance_spaces=tuple(instance_space) if instance_space is not None else None,
            )
        else:
            raise ToolkitValueError("Invalid combination of arguments.")
        cmd.run(
            lambda: cmd.instances(
                client=client,
                selector=selector,
                unlink=unlink,
                dry_run=dry_run,
                auto_yes=auto_yes,
                verbose=verbose,
            )
        )



from abc import ABC, abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Generic

from cognite.client.data_classes import DataSetList, filters
from cognite.client.data_classes._base import T_CogniteResourceList
from cognite.client.data_classes.aggregations import UniqueResult
from cognite.client.data_classes.assets import AssetProperty
from cognite.client.data_classes.data_modeling import (
    NodeApply,
    NodeApplyList,
    NodeOrEdgeData,
    SpaceApply,
    SpaceApplyList,
    ViewId,
)
from cognite.client.data_classes.documents import SourceFileProperty
from cognite.client.data_classes.events import EventProperty

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.apm_config_v1 import APMConfig, APMConfigList
from cognite_toolkit._cdf_tk.cruds import NodeCRUD, ResourceCRUD, SpaceCRUD
from cognite_toolkit._cdf_tk.exceptions import ToolkitMissingResourceError, ToolkitRequiredValueError
from cognite_toolkit._cdf_tk.utils import humanize_collection

from .data_model import CREATED_SOURCE_SYSTEM_VIEW_ID, SPACE, SPACE_SOURCE_VIEW_ID


@dataclass
class ResourceConfig:
    filestem: str
    data: dict[str, Any]


class MigrationCreator(ABC, Generic[T_CogniteResourceList]):
    """Base class for migration resources configurations that are created resources."""

    CRUD: type[ResourceCRUD]
    DISPLAY_NAME: str
    HAS_LINEAGE: bool = True

    def __init__(self, client: ToolkitClient) -> None:
        self.client = client

    @abstractmethod
    def create_resources(self) -> T_CogniteResourceList:
        raise NotImplementedError("Subclasses should implement this method")

    @abstractmethod
    def resource_configs(self, resources: T_CogniteResourceList) -> list[ResourceConfig]:
        raise NotImplementedError("Subclasses should implement this method")

    def store_lineage(self, resources: T_CogniteResourceList) -> int:
        """Store lineage information for the created resources.

        Args:
            resources: The list of created resources.
        """
        raise NotImplementedError("Subclasses should implement this method")


class InstanceSpaceCreator(MigrationCreator[SpaceApplyList]):
    """Creates instance spaces for migration."""

    CRUD = SpaceCRUD
    DISPLAY_NAME = "Instance Space"
    HAS_LINEAGE = True

    def __init__(
        self, client: ToolkitClient, datasets: DataSetList | None = None, data_set_external_ids: list[str] | None = None
    ) -> None:
        super().__init__(client)
        if sum([datasets is not None, data_set_external_ids is not None]) != 1:
            raise ValueError("Exactly one of datasets or data_set_external_ids must be provided.")
        self.data_set_external_ids = data_set_external_ids
        self.datasets = datasets or DataSetList([])

    def create_resources(self) -> SpaceApplyList:
        if self.data_set_external_ids is not None:
            self.datasets = self.client.data_sets.retrieve_multiple(external_ids=self.data_set_external_ids)

        if missing_external_ids := [ds.id for ds in self.datasets if ds.external_id is None]:
            raise ToolkitRequiredValueError(
                f"Cannot create instance spaces for datasets with missing external IDs: {humanize_collection(missing_external_ids)}"
            )

        return SpaceApplyList(
            [
                SpaceApply(
                    # This is checked above
                    space=dataset.external_id,  # type: ignore[arg-type]
                    name=dataset.name,
                    description=dataset.description,
                )
                for dataset in self.datasets
            ]
        )

    def resource_configs(self, resources: SpaceApplyList) -> list[ResourceConfig]:
        return [ResourceConfig(filestem=space.space, data=space.dump()) for space in resources]

    def store_lineage(self, resources: SpaceApplyList) -> int:
        data_set_by_external_id = {ds.external_id: ds for ds in self.datasets}
        nodes = NodeApplyList(
            [
                NodeApply(
                    space=SPACE.space,
                    external_id=space.space,
                    sources=[
                        NodeOrEdgeData(
                            source=SPACE_SOURCE_VIEW_ID,
                            properties={
                                "instanceSpace": space.space,
                                "dataSetId": data_set_by_external_id[space.space].id,
                                "dataSetExternalId": data_set_by_external_id[space.space].external_id,
                            },
                        )
                    ],
                )
                for space in resources
            ]
        )
        res = self.client.data_modeling.instances.apply(nodes)
        return len(res.nodes)


class SourceSystemCreator(MigrationCreator[NodeApplyList]):
    CRUD = NodeCRUD
    DISPLAY_NAME = "Source System"
    COGNITE_SOURCE_SYSTEM_VIEW_ID = ViewId("cdf_cdm", "CogniteSourceSystem", "v1")

    def __init__(
        self,
        client: ToolkitClient,
        instance_space: str,
        data_set_external_id: str | None = None,
        hierarchy: str | None = None,
    ) -> None:
        super().__init__(client)
        if sum([data_set_external_id is not None, hierarchy is not None]) != 1:
            raise ValueError("Exactly one of data_set_external_id or hierarchy must be provided.")
        self._instance_space = instance_space
        self.data_set_external_id = data_set_external_id
        self.hierarchy = hierarchy

    def create_resources(self) -> NodeApplyList:
        seen: set[str] = set()
        nodes = NodeApplyList([])
        for source in self._lookup_sources():
            source_str = source.value
            if not isinstance(source_str, str) or source_str in seen:
                continue
            seen.add(source_str)
            nodes.append(
                NodeApply(
                    space=self._instance_space,
                    external_id=source_str,
                    sources=[
                        NodeOrEdgeData(source=self.COGNITE_SOURCE_SYSTEM_VIEW_ID, properties={"name": source_str}),
                        NodeOrEdgeData(source=CREATED_SOURCE_SYSTEM_VIEW_ID, properties={"source": source_str}),
                    ],
                )
            )
        return nodes

    def resource_configs(self, resources: NodeApplyList) -> list[ResourceConfig]:
        output: list[ResourceConfig] = []
        for node in resources:
            copy = NodeApply(
                space=node.space,
                external_id=node.external_id,
                # We remove the lineage source as this is not expected to be part of the governed
                # SourceSystem.
                sources=[s for s in node.sources if s.source != CREATED_SOURCE_SYSTEM_VIEW_ID],
            )
            output.append(ResourceConfig(filestem=node.external_id, data=copy.dump()))
        return output

    def _lookup_sources(self) -> Iterable[UniqueResult]:
        yield from self.client.assets.aggregate_unique_values(AssetProperty.source, filter=self._simple_filter)
        yield from self.client.events.aggregate_unique_values(property=EventProperty.source, filter=self._simple_filter)
        yield from self.client.documents.aggregate_unique_values(
            SourceFileProperty.source, filter=self._advanced_filter, limit=1000
        )

    @cached_property
    def _simple_filter(self) -> dict[str, Any] | None:
        if self.data_set_external_id is not None:
            return {"dataSetIds": [{"externalId": self.data_set_external_id}]}
        if self.hierarchy is not None:
            return {"assetSubtreeIds": [{"externalId": self.hierarchy}]}
        return None

    @cached_property
    def _advanced_filter(self) -> filters.Filter:
        if self.data_set_external_id is not None:
            data_set_id = self.client.lookup.data_sets.id(self.data_set_external_id)
            if data_set_id is None:
                raise ToolkitMissingResourceError(f"Data set with external ID '{self.data_set_external_id}' not found.")
            return filters.Equals(SourceFileProperty.data_set_id, data_set_id)
        if self.hierarchy is not None:
            return filters.InAssetSubtree(SourceFileProperty.asset_external_ids, [self.hierarchy])
        else:
            raise ValueError("This should not happen.")

    def store_lineage(self, resources: NodeApplyList) -> int:
        # We already store lineage when creating the resources.
        return len(resources)


class InfieldV2ConfigCreator(MigrationCreator[NodeApplyList]):
    CRUD = NodeCRUD
    DISPLAY_NAME = "Infield V2 Configuration"
    HAS_LINEAGE = False

    def create_resources(self) -> NodeApplyList:
        apm_config_nodes = self.client.data_modeling.instances.list(instance_type="node", sources=APMConfig.view_id)
        apm_config = APMConfigList.from_nodes(apm_config_nodes)

        new_config_nodes = NodeApplyList([])
        for config in apm_config:
            new_config = self._create_infield_v2_config(config)
            new_config_nodes.append(new_config)
        return new_config_nodes

    def resource_configs(self, resources: NodeApplyList) -> list[ResourceConfig]:
        return [ResourceConfig(filestem=node.external_id, data=node.dump()) for node in resources]

    def _create_infield_v2_config(self, config: APMConfig) -> NodeApply:
        raise NotImplementedError("To be implemented")



import re
import tempfile
import textwrap
from pathlib import Path

from cognite.client.data_classes import UserProfile
from cognite.client.exceptions import CogniteAPIError
from rich import print
from rich.panel import Panel

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands import AuthCommand, BuildCommand, DeployCommand, ModulesCommand
from cognite_toolkit._cdf_tk.constants import _RUNNING_IN_BROWSER, MODULES
from cognite_toolkit._cdf_tk.cruds import CRUDS_BY_FOLDER_NAME
from cognite_toolkit._cdf_tk.exceptions import AuthenticationError
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables


class CogniteToolkitDemo:
    def __init__(self) -> None:
        if _RUNNING_IN_BROWSER:
            try:
                self._client = ToolkitClient()
            except CogniteAPIError as e:
                raise AuthenticationError(f"Failed to initialize CogniteClient in browser: {e}")
        else:
            self._client = EnvironmentVariables.create_from_environment().get_client()
        self._env_vars: EnvironmentVariables | None = None
        print(
            Panel(
                textwrap.dedent("""
        This is a demo version of the Cognite Toolkit.

        It is intended to demonstrate the content and capabilities of the Toolkit.
        It is not intended to be used in production or development environments.

        To learn more about the Cognite Toolkit, visit https://docs.cognite.com/cdf/deploy/cdf_toolkit/.
        """)
            )
        )

    @property
    def _tmp_path(self) -> Path:
        return Path(tempfile.gettempdir()).resolve()

    @property
    def _build_dir(self) -> Path:
        build_path = self._tmp_path / "cognite-toolkit-build"
        build_path.mkdir(exist_ok=True)
        return build_path

    @property
    def _organization_dir(self) -> Path:
        organization_path = self._tmp_path / "cognite-toolkit-organization"
        organization_path.mkdir(exist_ok=True)
        return organization_path

    def quickstart(
        self, company_prefix: str | None, client_id: str | None = None, client_secret: str | None = None
    ) -> None:
        if sum([client_id is None, client_secret is None]) == 1:
            raise ValueError("Both client_id and client_secret must be provided or neither.")
        if company_prefix:
            self._verify_company_prefix(company_prefix)
        print(Panel("Running Toolkit QuickStart..."))
        user = self._client.iam.user_profiles.me()
        if client_id is None and client_secret is None:
            print("Client ID and secret not provided. Assuming user has all the necessary permissions.")
            self._init_build_deploy(user, company_prefix)
            return

        group_id: int | None = None
        try:
            # Lookup user ID to add user ID to the group to run the workflow
            auth = AuthCommand()
            auth_result = auth.verify(
                self._client,
                dry_run=False,
                no_prompt=True,
                demo_principal=client_id,
            )
            group_id = auth_result.toolkit_group_id
            if auth_result.function_status is None:
                print(Panel("Unknown function status. If the demo fails, please check that functions are activated"))
            elif auth_result.function_status == "requested":
                print(
                    Panel(
                        "Function status is requested. Please wait for the function status to be activated before running the demo."
                    )
                )
                return
            elif auth_result.function_status == "inactive":
                print(Panel("Function status is inactive. Cannot run demo without functions."))
                return

            print("Switching to the demo service principal...")
            cluster = self._client.config.cdf_cluster
            if cluster is None:
                raise ValueError("CDF_CLUSTER is not set in the environment.")
            self._env_vars = EnvironmentVariables(
                PROVIDER="cdf",
                LOGIN_FLOW="client_credentials",
                CDF_CLUSTER=cluster,
                CDF_PROJECT=self._client.config.project,
                IDP_CLIENT_ID=client_id,
                IDP_CLIENT_SECRET=client_secret,
            )
            self._client = self._env_vars.get_client()
            self._init_build_deploy(user, company_prefix)
        finally:
            if group_id is not None:
                self._client.iam.groups.delete(id=group_id)

    def _init_build_deploy(self, user: UserProfile, company_prefix: str | None = None) -> None:
        modules_cmd = ModulesCommand()
        modules_cmd.run(
            lambda: modules_cmd.init(
                organization_dir=self._organization_dir,
                user_select="quickstart",
                clean=True,
                user_download_data=True,
                user_environments=["dev"],
            )
        )
        config_yaml = self._organization_dir / "config.dev.yaml"
        config_raw = config_yaml.read_text()
        # Ensure the user can execute the workflow
        config_raw = config_raw.replace("<your user id>", user.user_identifier)
        # To avoid warnings about not set values
        config_raw = config_raw.replace("<not set>", "123456-to-be-replaced")
        config_raw = config_raw.replace("<my-project-dev>", self._client.config.project)
        if company_prefix is not None:
            config_raw = config_raw.replace("YourOrg", company_prefix)
        config_yaml.write_text(config_raw)

        # The Workflow trigger expects credentials to be set in the environment, so we delete it as
        # the user is expected to trigger the workflow manually.
        for workflow_trigger_file in (self._organization_dir / MODULES).rglob("*WorkflowTrigger.yaml"):
            workflow_trigger_file.unlink()

        build = BuildCommand()
        build.run(
            lambda: build.execute(
                build_dir=self._build_dir,
                organization_dir=self._organization_dir,
                selected=None,
                build_env_name="dev",
                verbose=False,
                client=self._client,
                on_error="raise",
                no_clean=False,
            )
        )

        deploy = DeployCommand()
        if self._env_vars is None:
            raise AuthenticationError("Environment variables not set.")

        deploy.run(
            lambda: deploy.deploy_build_directory(
                # MyPy fails to see th check above.
                env_vars=self._env_vars,  # type: ignore[arg-type]
                build_dir=self._build_dir,
                build_env_name="dev",
                dry_run=False,
                drop_data=False,
                drop=False,
                force_update=False,
                include=list(CRUDS_BY_FOLDER_NAME.keys()),
                verbose=False,
            )
        )

    def _verify_company_prefix(self, company_prefix: str) -> None:
        """Needs to comply with the regex for container and views ExternalID

        * View:      ^[a-zA-Z]([a-zA-Z0-9_]{0,253}[a-zA-Z0-9])?$
        * Container: ^[a-zA-Z]([a-zA-Z0-9_]{0,253}[a-zA-Z0-9])?$
        """
        if not re.match(r"^[a-zA-Z]", company_prefix):
            raise ValueError("The company prefix must start with a letter.")
        if not re.match(r"^[a-zA-Z]([a-zA-Z0-9_]{0,253}[a-zA-Z0-9])?$", company_prefix):
            invalid_chars = re.findall(r"[^a-zA-Z0-9_]", company_prefix)
            if invalid_chars:
                raise ValueError(
                    f"The company prefix contains invalid characters: {', '.join(invalid_chars)}. Only letters, numbers, and underscores are allowed."
                )
            if len(company_prefix) > 255:
                raise ValueError("The company prefix must be 255 characters or less.")
            raise ValueError("The company prefix is invalid.")
        return



from typing import Literal
from urllib.parse import urljoin

from cognite.client import ClientConfig
from cognite.client.credentials import CredentialProvider


class ToolkitClientConfig(ClientConfig):
    def __init__(
        self,
        client_name: str,
        project: str,
        credentials: CredentialProvider,
        api_subversion: str | None = None,
        base_url: str | None = None,
        max_workers: int | None = None,
        is_strict_validation: bool = True,
        headers: dict[str, str] | None = None,
        timeout: int | None = None,
        file_transfer_timeout: int | None = None,
        debug: bool = False,
    ) -> None:
        super().__init__(
            client_name=client_name,
            project=project,
            credentials=credentials,
            api_subversion=api_subversion,
            base_url=base_url,
            max_workers=max_workers,
            headers=headers,
            timeout=timeout,
            file_transfer_timeout=file_transfer_timeout,
            debug=debug,
        )
        self.is_strict_validation = is_strict_validation

    @classmethod
    def from_client_config(cls, config: ClientConfig, is_strict_validation: bool = True) -> "ToolkitClientConfig":
        return cls(
            client_name=config.client_name,
            project=config.project,
            credentials=config.credentials,
            api_subversion=config.api_subversion,
            base_url=config.base_url,
            max_workers=config.max_workers,
            headers=config.headers,
            timeout=config.timeout,
            file_transfer_timeout=config.file_transfer_timeout,
            debug=config.debug,
            is_strict_validation=is_strict_validation,
        )

    @property
    def cloud_provider(self) -> Literal["azure", "aws", "gcp", "unknown"]:
        cdf_cluster = self.cdf_cluster
        if cdf_cluster is None:
            return "unknown"
        elif cdf_cluster.startswith("az-") or cdf_cluster in {"azure-dev", "bluefield", "westeurope-1"}:
            return "azure"
        elif cdf_cluster.startswith("aws-") or cdf_cluster in {"orangefield"}:
            return "aws"
        elif cdf_cluster.startswith("gc-") or cdf_cluster in {
            "greenfield",
            "asia-northeast1-1",
            "cognitedata-development",
            "cognitedata-production",
        }:
            return "gcp"
        else:
            return "unknown"

    @property
    def is_private_link(self) -> bool:
        if "cognitedata.com" not in self.base_url:
            return False
        subdomain = self.base_url.split("cognitedata.com", maxsplit=1)[0]
        return "plink" in subdomain

    @property
    def base_api_url(self) -> str:
        return f"{self.base_url}/api/v1/projects/{self.project}"

    def create_api_url(self, endpoint: str) -> str:
        """Create a full API URL for the given endpoint.

        Args:
            endpoint (str): The API endpoint to append to the base URL.

        Returns:
            str: The full API URL.

        Examples:
            >>> config = ToolkitClientConfig(cluster="bluefield", project="my_project", ...)
            >>> config.create_api_url("/models/instances")
            "https://bluefield.cognitedata.com/api/v1/my_project/models/instances"
        """
        if not endpoint.startswith("/"):
            endpoint = f"/{endpoint}"
        return f"{self.base_api_url}{endpoint}"

    def create_app_url(self, endpoint: str) -> str:
        """Create a full App URL for the given endpoint.

        Args:
            endpoint (str): The App endpoint to append to the base URL.

        Returns:
            str: The full App URL.

        Examples:
            >>> config = ToolkitClientConfig(cluster="bluefield", project="my_project", ...)
            >>> config.create_app_url("/some/app/endpoint")
            "https://bluefield.cognitedata.com/apps/v1/projects/my_project/some/app/endpoint"
        """
        if not endpoint:
            raise ValueError("Endpoint must be a non-empty string")
        if not endpoint.startswith("/"):
            endpoint = f"/{endpoint}"
        base_path = f"/apps/v1/projects/{self.project}{endpoint}"
        return urljoin(self.base_url, base_path)



from typing import Any

from .base import BaseModelResource, ToolkitResource


class ObservationFeatureToggles(BaseModelResource):
    """Feature toggles for observations."""

    is_enabled: bool | None = None
    is_write_back_enabled: bool | None = None
    notifications_endpoint_external_id: str | None = None
    attachments_endpoint_external_id: str | None = None


class FeatureToggles(BaseModelResource):
    """Feature toggles for InField location configuration."""

    three_d: bool | None = None
    trends: bool | None = None
    documents: bool | None = None
    workorders: bool | None = None
    notifications: bool | None = None
    media: bool | None = None
    template_checklist_flow: bool | None = None
    workorder_checklist_flow: bool | None = None
    observations: ObservationFeatureToggles | None = None


class AccessManagement(BaseModelResource):
    """Access management configuration."""

    template_admins: list[str] | None = None  # list of CDF group external IDs
    checklist_admins: list[str] | None = None  # list of CDF group external IDs


class ResourceFilters(BaseModelResource):
    """Resource filters."""

    spaces: list[str] | None = None


class RootLocationDataFilters(BaseModelResource):
    """Data filters for root location."""

    general: ResourceFilters | None = None
    assets: ResourceFilters | None = None
    files: ResourceFilters | None = None
    timeseries: ResourceFilters | None = None


class DataExplorationConfig(BaseModelResource):
    """Properties for DataExplorationConfig node.

    Contains configuration for data exploration features:
    - observations: Observations feature configuration
    - activities: Activities configuration
    - documents: Document configuration
    - notifications: Notifications configuration
    - assets: Asset page configuration
    """

    space: str | None = None
    external_id: str | None = None

    observations: dict[str, Any] | None = None  # ObservationsConfigFeature
    activities: dict[str, Any] | None = None  # ActivitiesConfiguration
    documents: dict[str, Any] | None = None  # DocumentConfiguration
    notifications: dict[str, Any] | None = None  # NotificationsConfiguration
    assets: dict[str, Any] | None = None  # AssetPageConfiguration


class InfieldLocationConfigYAML(ToolkitResource):
    """Properties for InFieldLocationConfig node.

    Currently migrated fields:
    - root_location_external_id: Reference to the LocationFilterDTO external ID
    - feature_toggles: Feature toggles migrated from old configuration
    - rootAsset: Direct relation to the root asset (space and externalId)
    - app_instance_space: Application instance space from appDataInstanceSpace
    - access_management: Template and checklist admin groups (from templateAdmins and checklistAdmins)
    - disciplines: List of disciplines (from disciplines in FeatureConfiguration)
    - data_filters: Data filters for general, assets, files, and timeseries (from dataFilters in old configuration)
    - data_exploration_config: Direct relation to the DataExplorationConfig node (shared across all locations)
    """

    space: str
    external_id: str

    root_location_external_id: str | None = None
    feature_toggles: FeatureToggles | None = None
    app_instance_space: str | None = None
    access_management: AccessManagement | None = None
    data_filters: RootLocationDataFilters | None = None
    data_exploration_config: DataExplorationConfig | None = None



from collections.abc import Iterable
from datetime import datetime
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.workflow_trigger import WorkflowTriggerYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_workflow_trigger_test_cases() -> Iterable:
    # Missing required field: externalId, extra field 'name'
    yield pytest.param(
        {"name": "MyWorkflowTrigger"},
        {
            "Missing required field: 'authentication'",
            "Missing required field: 'externalId'",
            "Missing required field: 'triggerRule'",
            "Missing required field: 'workflowExternalId'",
            "Missing required field: 'workflowVersion'",
            "Unused field: 'name'",
        },
        id="Missing required fields ",
    )
    # Extra/unknown field
    yield pytest.param(
        {
            "externalId": "abc",
            "triggerRule": {},
            "workflowExternalId": "wf",
            "workflowVersion": "v1",
            "authentication": {},
            "foo": 123,
        },
        {
            "In authentication missing required field: 'clientId'",
            "In authentication missing required field: 'clientSecret'",
            "In field triggerRule invalid trigger rule data missing 'triggerType' key",
            "Unused field: 'foo'",
        },
        id="Extra field and missing triggerType in triggerRule",
    )
    # Invalid trigger type
    yield pytest.param(
        {
            "externalId": "abc",
            "triggerRule": {"triggerType": "notAType"},
            "workflowExternalId": "wf",
            "workflowVersion": "v1",
            "authentication": {"clientId": "id", "clientSecret": "secret"},
        },
        {"In field triggerRule invalid trigger type 'notAType'. Expected one of dataModeling or schedule"},
        id="Invalid triggerType value",
    )
    # Missing required triggerRule fields (e.g. for schedule)
    yield pytest.param(
        {
            "externalId": "abc",
            "triggerRule": {"triggerType": "schedule"},
            "workflowExternalId": "wf",
            "workflowVersion": "v1",
            "authentication": {"clientId": "id", "clientSecret": "secret"},
        },
        {"In triggerRule missing required field: 'cronExpression'"},
        id="Missing required field in schedule triggerRule",
    )
    # Wrong type for batch_size in dataModeling trigger
    yield pytest.param(
        {
            "externalId": "abc",
            "triggerRule": {
                "triggerType": "dataModeling",
                "dataModelingQuery": {},
                "batchSize": "notAnInt",
                "batchTimeout": 100,
            },
            "workflowExternalId": "wf",
            "workflowVersion": "v1",
            "authentication": {"clientId": "id", "clientSecret": "secret"},
        },
        {"In triggerRule.batchSize input should be a valid integer, unable to parse string as an integer"},
        id="Wrong type for batchSize in dataModeling trigger",
    )
    # Invalid value for batch_timeout (too low)
    yield pytest.param(
        {
            "externalId": "abc",
            "triggerRule": {
                "triggerType": "dataModeling",
                "dataModelingQuery": {},
                "batchSize": 100,
                "batchTimeout": 10,
            },
            "workflowExternalId": "wf",
            "workflowVersion": "v1",
            "authentication": {"clientId": "id", "clientSecret": "secret"},
        },
        {"In triggerRule.batchTimeout input should be greater than or equal to 60"},
        id="Invalid value for batchTimeout in dataModeling trigger",
    )
    # Invalid Json input
    yield pytest.param(
        {
            "externalId": "abc",
            "input": {"not_json": datetime(2025, 9, 1)},
            "triggerRule": {"triggerType": "schedule", "cronExpression": "* * * * *"},
            "workflowExternalId": "wf",
            "workflowVersion": "v1",
            "authentication": {"clientId": "id", "clientSecret": "secret"},
        },
        {"In input.dict.not_json input was not a valid JSON value"},
        id="Invalid Json input in dataModelingQuery",
    )


class TestWorkflowYAML:
    @pytest.mark.parametrize("data", list(find_resources("WorkflowTrigger")))
    def test_load_valid_workflow_file(self, data: dict[str, object]) -> None:
        loaded = WorkflowTriggerYAML.model_validate(data)

        dumped = loaded.model_dump(exclude_unset=True, by_alias=True)
        assert "authentication" in dumped
        # Secret is not dumped as per design, so we add it back for comparison
        dumped["authentication"]["clientSecret"] = data["authentication"]["clientSecret"]
        assert dumped == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_workflow_trigger_test_cases()))
    def test_invalid_workflow_trigger_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, WorkflowTriggerYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



from abc import ABC, abstractmethod
from pathlib import Path

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel

from cognite_toolkit._cdf_tk.constants import DATA_MANIFEST_STEM
from cognite_toolkit._cdf_tk.utils.file import safe_write, sanitize_filename, yaml_safe_dump
from cognite_toolkit._cdf_tk.utils.text import to_sentence_case
from cognite_toolkit._cdf_tk.utils.useful_types import JsonVal


class SelectorObject(BaseModel):
    """This is used as base class for all selector objects including nested ones."""

    model_config = ConfigDict(frozen=True, alias_generator=to_camel, populate_by_name=True)


class DataSelector(SelectorObject, ABC):
    """A selector gives instructions on what data to select from CDF.

    For example, for instances it can be a view or container, while for assets it can be a data set or asset subtree.

    Args:
        type: The type of selector.
        kind: The kind of data the selector is for (e.g., 'RawRows', 'Assets').
    """

    type: str
    kind: str

    def dump(self) -> dict[str, JsonVal]:
        return self.model_dump(by_alias=True)

    def dump_to_file(self, directory: Path) -> Path:
        """Dumps the selector to a YAML file in the specified directory.

        The filename is derived from the string representation of the selector.

        Args:
            directory: The directory where the YAML file will be saved.
        """

        filepath = directory / f"{sanitize_filename(str(self))}.{DATA_MANIFEST_STEM}.yaml"
        filepath.parent.mkdir(parents=True, exist_ok=True)
        safe_write(file=filepath, content=yaml_safe_dump(self.model_dump(mode="json", by_alias=True)), encoding="utf-8")
        return filepath

    @property
    @abstractmethod
    def group(self) -> str:
        """A string representing the group of the selector, used for organizing files.

        It is used when downloading to determine the subdirectory within the output directory.

        For example, for raw table, the group would be the database name, while the selector itself
        would be the table name.
        """
        raise NotImplementedError()

    @property
    def display_name(self) -> str:
        """A human-readable name for the selector."""
        return to_sentence_case(self.kind)

    def __str__(self) -> str:
        # We want to force subclasses to implement __str__
        raise NotImplementedError()



from collections import Counter
from collections.abc import Iterable
from pathlib import Path
from typing import Any, get_args

import pytest

from cognite_toolkit._cdf_tk.commands._migrate.selectors import AssetCentricMigrationSelector
from cognite_toolkit._cdf_tk.storageio import (
    AssetIO,
    ChartIO,
    DatapointsIO,
    InstanceIO,
    RawIO,
    StorageIO,
    get_upload_io,
)
from cognite_toolkit._cdf_tk.storageio.selectors import (
    AllChartsSelector,
    AssetCentricFileSelector,
    AssetSubtreeSelector,
    ChartOwnerSelector,
    DataPointsFileSelector,
    DataSelector,
    DataSetSelector,
    InstanceFileSelector,
    InstanceSpaceSelector,
    InstanceViewSelector,
    RawTableSelector,
    Selector,
    SelectorAdapter,
)
from cognite_toolkit._cdf_tk.utils import humanize_collection
from cognite_toolkit._cdf_tk.utils._auxiliary import get_concrete_subclasses
from cognite_toolkit._cdf_tk.utils.file import read_yaml_file


def example_selector_data() -> Iterable[tuple]:
    yield pytest.param(
        {"type": "rawTable", "table": {"dbName": "my_db", "tableName": "my_table"}},
        RawTableSelector,
        RawIO,
        RawIO.KIND,
        id="RawTableSelector",
    )
    yield pytest.param(
        {
            "type": "instanceView",
            "view": {"space": "my_space", "externalId": "my_view", "version": "v1"},
            "instanceType": "node",
            "instanceSpaces": ["space1", "space2"],
        },
        InstanceViewSelector,
        InstanceIO,
        InstanceIO.KIND,
        id="InstanceViewSelector",
    )
    yield pytest.param(
        {"type": "instanceFile", "datafile": "path/to/file.csv", "validateInstance": True},
        InstanceFileSelector,
        InstanceIO,
        InstanceIO.KIND,
        id="InstanceFileSelector",
    )
    yield pytest.param(
        {"type": "dataSet", "dataSetExternalId": "my_data_set", "kind": "Assets"},
        DataSetSelector,
        AssetIO,
        AssetIO.KIND,
        id="DataSetSelector",
    )
    yield pytest.param(
        {"type": "assetSubtree", "hierarchy": "root/child", "kind": "Assets"},
        AssetSubtreeSelector,
        AssetIO,
        AssetIO.KIND,
        id="AssetSubtreeSelector",
    )
    yield pytest.param(
        {"type": "chartOwner", "ownerId": "doctrino"},
        ChartOwnerSelector,
        ChartIO,
        ChartIO.KIND,
        id="ChartOwnerSelector",
    )
    yield pytest.param(
        {"type": "allCharts"},
        AllChartsSelector,
        ChartIO,
        ChartIO.KIND,
        id="AllChartSelector",
    )
    yield pytest.param(
        {"type": "assetFile", "datafile": "path/to/file.csv", "kind": "Assets"},
        AssetCentricFileSelector,
        AssetIO,
        AssetIO.KIND,
        id="AssetCentricFileSelector",
    )
    yield pytest.param(
        {
            "type": "instanceSpace",
            "instanceSpace": "myLocation",
            "instanceType": "node",
            "viewId": {
                "space": "my_space",
                "externalId": "MyView",  # Version is optional
            },
        },
        InstanceSpaceSelector,
        InstanceIO,
        InstanceIO.KIND,
        id="InstanceSpaceSelector",
    )
    yield pytest.param(
        {
            "type": "datapointsFile",
            "timestamp_column": "timestamp",
            "columns": [
                {
                    "column_type": "instance",
                    "column": "ts1",
                    "dtype": "numeric",
                    "space": "space1",
                    "external_id": "extId1",
                },
                {
                    "column_type": "externalId",
                    "column": "ts2",
                    "dtype": "string",
                    "external_id": "extId2",
                },
            ],
        },
        DataPointsFileSelector,
        DatapointsIO,
        DatapointsIO.KIND,
        id="DataPointsFileSelector",
    )


@pytest.fixture(scope="module")
def all_selectors() -> list[type[DataSelector]]:
    return get_concrete_subclasses(DataSelector)


class TestDataSelectors:
    """Test to ensure all data selectors are working as expected."""

    def test_all_selectors_in_union(self, all_selectors: list[type[DataSelector]]) -> None:
        # The migration selectors are not part of the Selector union, they
        # are only used for migration commands.
        migration_selectors = get_concrete_subclasses(AssetCentricMigrationSelector)
        all_union_selectors = get_args(Selector.__args__[0])
        missing = set(all_selectors) - set(all_union_selectors) - set(migration_selectors)
        assert not missing, (
            f"The following DataSelector subclasses are "
            f"missing from the Selector union: {humanize_collection([cls.__name__ for cls in missing])}"
        )

    def test_all_types_are_unique(self, all_selectors: list[type[DataSelector]]) -> None:
        types = Counter(cls.model_fields["type"].default for cls in all_selectors)
        duplicates = [t for t, count in types.items() if count > 1]
        assert not duplicates, f"The following DataSelector types are not unique: {humanize_collection(duplicates)}"

    def test_example_data_is_complete(self) -> None:
        migration_selectors = get_concrete_subclasses(AssetCentricMigrationSelector)
        # Migration selectors are not part of the Selector union, and are not
        # required to have example data here.
        all_selectors = [cls for cls in get_concrete_subclasses(DataSelector) if cls not in migration_selectors]
        example_types = {p.values[0]["type"] for p in example_selector_data()}
        all_types = {cls.model_fields["type"].default for cls in all_selectors}
        missing = all_types - example_types
        assert not missing, f"The following DataSelector types are missing example data: {humanize_collection(missing)}"

    @pytest.mark.parametrize("data,expected_selector,expected_io,kind", list(example_selector_data()))
    def test_selector_instance(
        self,
        data: dict[str, Any],
        expected_selector: type[DataSelector],
        expected_io: type[StorageIO],
        kind: str,
        tmp_path: Path,
    ) -> None:
        instance = SelectorAdapter.validate_python(data)

        # Assert correct type
        assert isinstance(instance, expected_selector), (
            f"Expected {expected_selector.__name__}, got {type(instance).__name__}"
        )
        # Assert __str__ is implemented
        assert str(instance), f"__str__ not implemented for {type(instance).__name__}"

        # Assert group is implemented
        assert instance.group, f"group property not implemented for {type(instance).__name__}"

        # Assert correct IO type
        assert get_upload_io(type(instance), kind) is expected_io, (
            f"Expected {expected_io.__name__} for selector {type(instance).__name__}, got {get_upload_io(type(instance), kind).__name__}"
        )

        # Assert selector is hashable
        assert isinstance(hash(instance), int), f"{type(instance).__name__} is not hashable"

        # Assert serialization/deserialization
        filepath = instance.dump_to_file(tmp_path)
        assert filepath.exists(), f"dump_to_file did not create file for {type(instance).__name__}"

        data = read_yaml_file(filepath)
        loaded = SelectorAdapter.validate_python(data)
        assert loaded.model_dump() == instance.model_dump()
        assert type(loaded) is type(instance)


class TestGetUploadIO:
    @pytest.mark.parametrize(
        "selector,path,expected_io",
        [
            pytest.param(
                RawTableSelector,
                Path(f"data.{RawIO.KIND}.csv"),
                RawIO,
                id="RawTableSelector with path",
            ),
            pytest.param(
                RawTableSelector,
                Path(f"data.{RawIO.KIND}.csv.gz"),
                RawIO,
                id="RawTableSelector with compressed data",
            ),
        ],
    )
    def test_get_upload_io_with_path(
        self, selector: type[DataSelector], path: Path, expected_io: type[StorageIO]
    ) -> None:
        assert get_upload_io(selector, path) == expected_io



from cognite.client.data_classes import Function

from cognite_toolkit._cdf_tk.client import ToolkitClient


class TestLookupFunctionsIds:
    def test_lookup_functions_id(self, toolkit_client: ToolkitClient, dummy_function: Function) -> None:
        function_id = toolkit_client.lookup.functions.id(dummy_function.external_id)

        assert function_id == dummy_function.id

    def test_lookup_functions_external_id(self, toolkit_client: ToolkitClient, dummy_function: Function) -> None:
        external_id = toolkit_client.lookup.functions.external_id(dummy_function.id)

        assert external_id == dummy_function.external_id

    def test_lookup_functions_external_id_not_found(self, toolkit_client: ToolkitClient) -> None:
        result = toolkit_client.lookup.functions.external_id(999999)

        assert result is None

    def test_lookup_functions_id_not_found_raise(self, toolkit_client: ToolkitClient) -> None:
        result = toolkit_client.lookup.functions.id("non_existent_function")

        assert result is None

    def test_lookup_functions_id_not_found(self, toolkit_client: ToolkitClient) -> None:
        result = toolkit_client.lookup.functions.id("", allow_empty=True)

        assert result == 0



from datetime import datetime, timezone
from unittest.mock import MagicMock

import pytest
from cognite.client._api.iam import IAMAPI, ComparableCapability
from cognite.client.data_classes.capabilities import (
    Capability,
    DataModelInstancesAcl,
    DataModelsAcl,
    ProjectCapabilityList,
)
from cognite.client.data_classes.data_modeling import DataModel, DataModelList, NodeList

from cognite_toolkit._cdf_tk.client.data_classes.canvas import Canvas, IndustrialCanvas, IndustrialCanvasApply
from cognite_toolkit._cdf_tk.client.data_classes.migration import InstanceSource
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.commands import MigrationCanvasCommand
from cognite_toolkit._cdf_tk.commands._migrate.data_model import (
    COGNITE_MIGRATION_MODEL,
    INSTANCE_SOURCE_VIEW_ID,
    MODEL_ID,
    RESOURCE_VIEW_MAPPING_VIEW_ID,
)
from cognite_toolkit._cdf_tk.exceptions import AuthenticationError, ToolkitMigrationError
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning, ToolkitWarning


class TestMigrationCanvasCommand:
    def test_migrate_canvas_happy_path(
        self, asset_centric_canvas: tuple[IndustrialCanvas, NodeList[InstanceSource]]
    ) -> None:
        command = MigrationCanvasCommand(silent=True)
        canvas, instance_sources = asset_centric_canvas

        with monkeypatch_toolkit_client() as client:
            client.iam.verify_capabilities.return_value = []
            client.data_modeling.data_models.retrieve.return_value = [COGNITE_MIGRATION_MODEL]
            client.canvas.industrial.retrieve.return_value = canvas
            client.migration.instance_source.retrieve.return_value = instance_sources

            command.migrate_canvas(client, external_ids=["my_canvas"], dry_run=False, verbose=True)

            client.canvas.industrial.update.assert_called_once()
            update = client.canvas.industrial.update.call_args[0][0]
            assert isinstance(update, IndustrialCanvasApply)
            client.canvas.industrial.create.assert_called_once()
            backup = client.canvas.industrial.create.call_args[0][0]
            assert isinstance(backup, IndustrialCanvasApply)

        assert len(update.fdm_instance_container_references) == len(canvas.container_references)
        assert len(backup.fdm_instance_container_references) == 0

    def test_migrate_canvas_missing(self) -> None:
        command = MigrationCanvasCommand(silent=True)

        with monkeypatch_toolkit_client() as client:
            client.iam.verify_capabilities.return_value = []
            client.data_modeling.data_models.retrieve.return_value = [COGNITE_MIGRATION_MODEL]
            client.canvas.industrial.retrieve.return_value = None
            command.migrate_canvas(client, external_ids=["non-existing"], dry_run=False, verbose=True)
        assert len(command.warning_list) == 1
        warning = command.warning_list[0]
        assert isinstance(warning, ToolkitWarning)
        assert "Canvas with external ID 'non-existing' not found." in str(warning)

    def test_migrate_canvas_missing_instance_source(
        self, asset_centric_canvas: tuple[IndustrialCanvas, NodeList[InstanceSource]]
    ) -> None:
        command = MigrationCanvasCommand(silent=True)
        canvas, _ = asset_centric_canvas

        with monkeypatch_toolkit_client() as client:
            client.iam.verify_capabilities.return_value = []
            client.data_modeling.data_models.retrieve.return_value = [COGNITE_MIGRATION_MODEL]
            client.canvas.industrial.retrieve.return_value = canvas
            client.migration.instance_source.retrieve.return_value = NodeList[InstanceSource]([])

            command.migrate_canvas(client, external_ids=[canvas.canvas.external_id], dry_run=False, verbose=True)

        assert len(command.warning_list) == 1
        warning = command.warning_list[0]
        assert isinstance(warning, HighSeverityWarning)
        assert "Canvas 'Asset-centric1' has references to resources that are not been migrated" in str(warning)

    def test_migrate_canvas_no_asset_centric_references(
        self, asset_centric_canvas: tuple[IndustrialCanvas, NodeList[InstanceSource]]
    ) -> None:
        command = MigrationCanvasCommand(silent=True)
        canvas = IndustrialCanvas(
            canvas=Canvas(
                "canvasSpace",
                "canvasExternalId",
                1,
                1,
                1,
                name="MyCanvas",
                created_by="me",
                updated_at=datetime(2020, 1, 1, tzinfo=timezone.utc),
                updated_by="me",
            )
        )

        with monkeypatch_toolkit_client() as client:
            client.iam.verify_capabilities.return_value = []
            client.data_modeling.data_models.retrieve.return_value = [COGNITE_MIGRATION_MODEL]
            client.canvas.industrial.retrieve.return_value = canvas

            command.migrate_canvas(client, external_ids=[canvas.canvas.external_id], dry_run=False, verbose=True)

        assert len(command.warning_list) == 1
        warning = command.warning_list[0]
        assert isinstance(warning, ToolkitWarning)
        assert "Canvas with name 'MyCanvas' does not have any asset-centric references." in str(warning)

    def test_validate_access(self) -> None:
        expected_missing = [
            DataModelInstancesAcl(
                actions=[
                    DataModelInstancesAcl.Action.Read,
                ],
                scope=DataModelInstancesAcl.Scope.SpaceID(["my_instance_space"]),
            ),
            DataModelInstancesAcl(
                actions=[
                    DataModelInstancesAcl.Action.Write,
                ],
                scope=DataModelInstancesAcl.Scope.SpaceID(["my_instance_space"]),
            ),
            DataModelInstancesAcl(
                actions=[
                    DataModelInstancesAcl.Action.Write_Properties,
                ],
                scope=DataModelInstancesAcl.Scope.SpaceID(["my_instance_space"]),
            ),
            DataModelsAcl(actions=[DataModelsAcl.Action.Read], scope=DataModelsAcl.Scope.SpaceID(["dummy_space"])),
        ]

        with monkeypatch_toolkit_client() as client:
            existing_capabilities = ProjectCapabilityList([], cognite_client=client)

            def verify_capabilities(desired_capabilities: ComparableCapability) -> list[Capability]:
                return IAMAPI.compare_capabilities(
                    existing_capabilities,
                    desired_capabilities,
                )

            client.iam.verify_capabilities = verify_capabilities
            with pytest.raises(AuthenticationError) as exc_info:
                MigrationCanvasCommand.validate_access(client, ["my_instance_space"], ["dummy_space"])
        error = exc_info.value
        assert isinstance(error, AuthenticationError)
        assert "Missing required capabilities" in error.args[0]
        assert error.args[1] == expected_missing

    def test_validate_migration_model_available(self) -> None:
        with monkeypatch_toolkit_client() as client:
            client.data_modeling.data_models.retrieve.return_value = DataModelList([])
            with pytest.raises(ToolkitMigrationError):
                MigrationCanvasCommand.validate_migration_model_available(client)

    def test_validate_migration_model_available_multiple_models(self) -> None:
        """Test that multiple models raises an error."""
        with monkeypatch_toolkit_client() as client:
            # Create mock models with the expected MODEL_ID
            model1 = MagicMock(spec=DataModel)
            model1.as_id.return_value = MODEL_ID
            model2 = MagicMock(spec=DataModel)
            model2.as_id.return_value = MODEL_ID

            client.data_modeling.data_models.retrieve.return_value = DataModelList([model1, model2])

            with pytest.raises(ToolkitMigrationError) as exc_info:
                MigrationCanvasCommand.validate_migration_model_available(client)

            assert "Multiple migration models" in str(exc_info.value)

    def test_validate_migration_model_available_missing_views(self) -> None:
        """Test that a model with missing views raises an error."""
        with monkeypatch_toolkit_client() as client:
            model = MagicMock(spec=DataModel)
            model.as_id.return_value = MODEL_ID
            # Model has views but missing the required ones
            model.views = [INSTANCE_SOURCE_VIEW_ID]  # Missing VIEW_SOURCE_VIEW_ID

            client.data_modeling.data_models.retrieve.return_value = DataModelList([model])

            with pytest.raises(ToolkitMigrationError, match=r"Invalid migration model. Missing views"):
                MigrationCanvasCommand.validate_migration_model_available(client)

    def test_validate_migration_model_available_success(self) -> None:
        """Test that a valid model with all required views succeeds."""
        with monkeypatch_toolkit_client() as client:
            # Mocking the migration Model to get a response format of the model.
            # An alternative would be to write a conversion of write -> read format of the model
            # which is a significant amount of logic.
            model = MagicMock(spec=DataModel)
            model.as_id.return_value = MODEL_ID
            # Model has all required views
            model.views = [INSTANCE_SOURCE_VIEW_ID, RESOURCE_VIEW_MAPPING_VIEW_ID]

            client.data_modeling.data_models.retrieve.return_value = DataModelList([model])

            # Should not raise any exception
            MigrationCanvasCommand.validate_migration_model_available(client)

            client.data_modeling.data_models.retrieve.assert_called_once_with([MODEL_ID], inline_views=False)



from collections.abc import Hashable
from pathlib import Path
from unittest.mock import MagicMock

import pytest
from _pytest.monkeypatch import MonkeyPatch
from cognite.client.data_classes import Group, GroupWrite

from cognite_toolkit._cdf_tk.client.data_classes.raw import RawDatabase, RawTable
from cognite_toolkit._cdf_tk.cruds import (
    DataSetsCRUD,
    ExtractionPipelineCRUD,
    GroupAllScopedCRUD,
    GroupCRUD,
    GroupResourceScopedCRUD,
    RawDatabaseCRUD,
    RawTableCRUD,
    ResourceCRUD,
    ResourceWorker,
    SpaceCRUD,
)
from cognite_toolkit._cdf_tk.exceptions import ToolkitWrongResourceError
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables
from tests.data import LOAD_DATA
from tests.test_unit.approval_client import ApprovalToolkitClient


class TestGroupLoader:
    def test_load_all_scoped_only(self, env_vars_with_client: EnvironmentVariables, monkeypatch: MonkeyPatch):
        loader = GroupAllScopedCRUD.create_loader(env_vars_with_client.get_client())
        raw_list = loader.load_resource_file(
            LOAD_DATA / "auth" / "1.my_group_unscoped.yaml", env_vars_with_client.dump()
        )
        loaded = loader.load_resource(raw_list[0], is_dry_run=False)
        assert loaded.name == "unscoped_group_name"

        raw_list = loader.load_resource_file(LOAD_DATA / "auth" / "1.my_group_scoped.yaml", env_vars_with_client.dump())
        with pytest.raises(ToolkitWrongResourceError):
            loader.load_resource(raw_list[0], is_dry_run=False)

    def test_load_resource_scoped_only(self, env_vars_with_client: EnvironmentVariables, monkeypatch: MonkeyPatch):
        loader = GroupResourceScopedCRUD.create_loader(env_vars_with_client.get_client())
        with pytest.raises(ToolkitWrongResourceError):
            raw_list = loader.load_resource_file(
                LOAD_DATA / "auth" / "1.my_group_unscoped.yaml", env_vars_with_client.dump()
            )
            loader.load_resource(raw_list[0], is_dry_run=False)

        raw_list = loader.load_resource_file(LOAD_DATA / "auth" / "1.my_group_scoped.yaml", env_vars_with_client.dump())
        loaded = loader.load_resource(raw_list[0], is_dry_run=False)
        assert loaded.name == "scoped_group_name"
        assert len(loaded.capabilities) == 4

        caps = {str(type(element).__name__): element for element in loaded.capabilities}

        assert all(isinstance(item, int) for item in caps["DataSetsAcl"].scope.ids)
        assert all(isinstance(item, int) for item in caps["AssetsAcl"].scope.ids)
        assert all(isinstance(item, int) for item in caps["ExtractionConfigsAcl"].scope.ids)
        assert caps["SessionsAcl"].scope._scope_name == "all"

    def test_load_group_list_resource_scoped_only(
        self, env_vars_with_client: EnvironmentVariables, monkeypatch: MonkeyPatch
    ):
        loader = GroupResourceScopedCRUD.create_loader(env_vars_with_client.get_client())
        raw_list = loader.load_resource_file(
            LOAD_DATA / "auth" / "1.my_group_list_combined.yaml", env_vars_with_client.dump()
        )
        loaded = loader.load_resource(raw_list[0], is_dry_run=False)

        assert isinstance(loaded, GroupWrite)
        assert loaded.name == "scoped_group_name"

    def test_load_group_list_all_scoped_only(
        self, env_vars_with_client: EnvironmentVariables, monkeypatch: MonkeyPatch
    ):
        loader = GroupAllScopedCRUD.create_loader(env_vars_with_client.get_client())
        raw_list = loader.load_resource_file(
            LOAD_DATA / "auth" / "1.my_group_list_combined.yaml", env_vars_with_client.dump()
        )
        loaded = loader.load_resource(raw_list[1], is_dry_run=False)

        assert isinstance(loaded, GroupWrite)
        assert loaded.name == "unscoped_group_name"

    def test_unchanged_new_group(
        self,
        env_vars_with_client: EnvironmentVariables,
        toolkit_client_approval: ApprovalToolkitClient,
        monkeypatch: MonkeyPatch,
    ) -> None:
        loader = GroupResourceScopedCRUD.create_loader(env_vars_with_client.get_client())
        raw_list = loader.load_resource_file(LOAD_DATA / "auth" / "1.my_group_scoped.yaml", env_vars_with_client.dump())
        loaded = loader.load_resource(raw_list[0], is_dry_run=False)

        # Simulate that one group is is already in CDF
        toolkit_client_approval.append(
            Group,
            [
                Group(
                    id=123,
                    name=loaded.name,
                    source_id=loaded.source_id,
                    capabilities=loaded.capabilities,
                    metadata=loaded.metadata,
                    is_deleted=False,
                )
            ],
        )

        new_file = MagicMock(spec=Path)
        new_file.read_text.return_value = GroupWrite(
            name="new_group", source_id="123", capabilities=loaded.capabilities
        ).dump_yaml()
        worker = ResourceWorker(loader, "deploy")
        resources = worker.prepare_resources(
            [
                LOAD_DATA / "auth" / "1.my_group_scoped.yaml",
                new_file,
            ]
        )
        assert {
            "create": len(resources.to_create),
            "change": len(resources.to_update),
            "delete": len(resources.to_delete),
            "unchanged": len(resources.unchanged),
        } == {"create": 1, "change": 0, "delete": 0, "unchanged": 1}

    def test_upsert_group(
        self,
        env_vars_with_client: EnvironmentVariables,
        toolkit_client_approval: ApprovalToolkitClient,
        monkeypatch: MonkeyPatch,
    ):
        loader = GroupResourceScopedCRUD.create_loader(env_vars_with_client.get_client())
        raw_list = loader.load_resource_file(LOAD_DATA / "auth" / "1.my_group_scoped.yaml", env_vars_with_client.dump())
        loaded = loader.load_resource(raw_list[0], is_dry_run=False)

        # Simulate that the group is is already in CDF, but with fewer capabilities
        # Simulate that one group is is already in CDF
        toolkit_client_approval.append(
            Group,
            [
                Group(
                    id=123,
                    name=loaded.name,
                    source_id=loaded.source_id,
                    capabilities=loaded.capabilities[0:1],
                    metadata=loaded.metadata,
                    is_deleted=False,
                )
            ],
        )

        # group exists, no changes
        worker = ResourceWorker(loader, "deploy")
        resources = worker.prepare_resources([LOAD_DATA / "auth" / "1.my_group_scoped.yaml"])

        assert {
            "create": len(resources.to_create),
            "change": len(resources.to_update),
            "delete": len(resources.to_delete),
            "unchanged": len(resources.unchanged),
        } == {"create": 0, "change": 1, "delete": 0, "unchanged": 0}

    @pytest.mark.parametrize(
        "item, expected",
        [
            pytest.param(
                {"capabilities": [{"dataModelsAcl": {"scope": {"spaceIdScope": {"spaceIds": ["space1", "space2"]}}}}]},
                [(SpaceCRUD, "space1"), (SpaceCRUD, "space2")],
                id="SpaceId scope",
            ),
            pytest.param(
                {"capabilities": [{"timeSeriesAcl": {"scope": {"datasetScope": {"ids": ["ds_dataset1"]}}}}]},
                [
                    (DataSetsCRUD, "ds_dataset1"),
                ],
                id="Dataset scope",
            ),
            pytest.param(
                {
                    "capabilities": [
                        {"extractionRunsAcl": {"scope": {"extractionPipelineScope": {"ids": ["ex_my_extraction"]}}}}
                    ]
                },
                [
                    (ExtractionPipelineCRUD, "ex_my_extraction"),
                ],
                id="Extraction pipeline scope",
            ),
            pytest.param(
                {"capabilities": [{"rawAcl": {"scope": {"tableScope": {"dbsToTables": {"my_db": ["my_table"]}}}}}]},
                [
                    (RawDatabaseCRUD, RawDatabase("my_db")),
                    (RawTableCRUD, RawTable("my_db", "my_table")),
                ],
                id="Table scope",
            ),
            pytest.param(
                {"capabilities": [{"datasetsAcl": {"scope": {"idscope": {"ids": ["ds_my_dataset"]}}}}]},
                [
                    (DataSetsCRUD, "ds_my_dataset"),
                ],
                id="ID scope dataset",
            ),
            pytest.param(
                {"capabilities": [{"extractionPipelinesAcl": {"scope": {"idscope": {"ids": ["ex_my_extraction"]}}}}]},
                [
                    (ExtractionPipelineCRUD, "ex_my_extraction"),
                ],
                id="ID scope extractionpipline ",
            ),
        ],
    )
    def test_get_dependent_items(self, item: dict, expected: list[tuple[type[ResourceCRUD], Hashable]]) -> None:
        actual_dependent_items = GroupCRUD.get_dependent_items(item)

        assert list(actual_dependent_items) == expected

    def test_unchanged_new_group_without_metadata(
        self,
        env_vars_with_client: EnvironmentVariables,
        toolkit_client_approval: ApprovalToolkitClient,
        monkeypatch: MonkeyPatch,
    ) -> None:
        loader = GroupAllScopedCRUD.create_loader(env_vars_with_client.get_client())
        local_group = """name: gp_no_metadata
sourceId: 123
capabilities:
- assetsAcl:
    actions:
    - READ
    scope:
      all: {}
"""
        cdf_group = Group.load("""name: gp_no_metadata
sourceId: 123
capabilities:
- assetsAcl:
    actions:
    - READ
    scope:
      all: {}
metadata: {}
id: 3760258445038144
isDeleted: false
deletedTime: -1
""")

        # Simulate that one group is is already in CDF
        toolkit_client_approval.append(
            Group,
            [cdf_group],
        )
        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = local_group

        worker = ResourceWorker(loader, "deploy")
        resources = worker.prepare_resources([filepath])
        assert {
            "create": len(resources.to_create),
            "change": len(resources.to_update),
            "delete": len(resources.to_delete),
            "unchanged": len(resources.unchanged),
        } == {"create": 0, "change": 0, "delete": 0, "unchanged": 1}

    def test_unchanged_group_raw_acl_table_scoped(
        self,
        env_vars_with_client: EnvironmentVariables,
        toolkit_client_approval: ApprovalToolkitClient,
        monkeypatch: MonkeyPatch,
    ) -> None:
        loader = GroupResourceScopedCRUD.create_loader(env_vars_with_client.get_client())
        local_group = """name: gp_raw_acl_table_scoped
sourceId: '123'
capabilities:
- rawAcl:
   actions:
   - READ
   scope:
     tableScope:
       dbsToTables:
         'db_name':
           - labels
        """
        cdf_group = Group.load("""name: gp_raw_acl_table_scoped
sourceId: '123'
capabilities:
- rawAcl:
    actions:
    - READ
    scope:
      tableScope:
        dbsToTables:
          'db_name':
            tables:
            - labels
metadata: {}
id: 3760258445038144
isDeleted: false
deletedTime: -1
        """)

        # Simulate that one group is is already in CDF
        toolkit_client_approval.append(
            Group,
            [cdf_group],
        )
        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = local_group

        worker = ResourceWorker(loader, "deploy")
        resources = worker.prepare_resources([filepath])
        assert {
            "create": len(resources.to_create),
            "change": len(resources.to_update),
            "delete": len(resources.to_delete),
            "unchanged": len(resources.unchanged),
        } == {"create": 0, "change": 0, "delete": 0, "unchanged": 1}



import shutil
from collections import defaultdict
from collections.abc import Callable
from pathlib import Path
from typing import Any

from rich import print
from rich.table import Table

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.exceptions import AuthenticationError, ToolkitValueError
from cognite_toolkit._cdf_tk.tk_warnings import LowSeverityWarning
from cognite_toolkit._cdf_tk.utils import read_yaml_file, safe_write
from cognite_toolkit._cdf_tk.utils.file import safe_rmtree, yaml_safe_dump


class ImportTransformationCLI(ToolkitCommand):
    def __init__(
        self,
        get_client: Callable[[], ToolkitClient] | None = None,
        print_warning: bool = True,
        skip_tracking: bool = False,
    ):
        super().__init__(print_warning, skip_tracking)
        self._dataset_external_id_by_id: dict[int, str] = {}
        # We only initialize the client if we need to look up dataset ids.
        self._client: ToolkitClient | None = None
        self._get_client = get_client

    def execute(
        self,
        source: Path,
        destination: Path,
        overwrite: bool,
        flatten: bool,
        clean: bool,
        verbose: bool = False,
    ) -> None:
        # Manifest files are documented at
        # https://cognite-transformations-cli.readthedocs-hosted.com/en/latest/quickstart.html#transformation-manifest
        if source.is_file() and source.suffix in {".yaml", ".yml"}:
            yaml_files = [source]
        elif source.is_file():
            raise ToolkitValueError(f"File {source} is not a YAML file.")
        elif source.is_dir():
            yaml_files = list(source.rglob("*.yaml")) + list(source.rglob("*.yml"))
        else:
            raise ToolkitValueError(f"Source {source} is not a file or directory.")

        if not yaml_files:
            self.warn(LowSeverityWarning("No YAML files found in the source directory."))
            return None

        count_by_resource_type: dict[str, int] = defaultdict(int)
        for yaml_file in yaml_files:
            data = self._load_file(yaml_file)
            if data is None:
                continue

            # The convert schedule and notifications pop off the schedule and notifications
            # keys from the transformation
            schedule = self._convert_schedule(data, data["externalId"], yaml_file)
            notifications = self._convert_notifications(data, data["externalId"], yaml_file)
            transformation, source_query_relative_path = self._convert_transformation(data, yaml_file)
            source_query_path = yaml_file.parent / source_query_relative_path if source_query_relative_path else None

            if source_query_path and not source_query_path.exists():
                raise ToolkitValueError(
                    f"Query file {source_query_path.as_posix()!r} does not exist. "
                    f"This is referenced in {yaml_file.as_posix()!r}."
                )

            if flatten:
                destination_folder = destination
            else:
                destination_folder = destination / yaml_file.relative_to(source).parent
            destination_folder.mkdir(parents=True, exist_ok=True)

            destination_transformation = destination_folder / f"{yaml_file.stem}.Transformation.yaml"
            if not overwrite and destination_transformation.exists():
                self.warn(LowSeverityWarning(f"File already exists at {destination_transformation}. Skipping."))
                continue
            safe_write(destination_transformation, yaml_safe_dump(transformation))
            if source_query_path is not None:
                destination_query_path = destination_folder / f"{destination_transformation.stem}.sql"
                shutil.copyfile(source_query_path, destination_query_path)

            if schedule is not None:
                destination_schedule = destination_folder / f"{yaml_file.stem}.Schedule.yaml"
                safe_write(destination_schedule, yaml_safe_dump(schedule))
            if notifications:
                destination_notification = destination_folder / f"{yaml_file.stem}.Notification.yaml"
                safe_write(destination_notification, yaml_safe_dump(notifications))
            if verbose:
                print(f"Imported {yaml_file} to {destination_folder}.")
            count_by_resource_type["transformation"] += 1
            count_by_resource_type["schedule"] += 1 if schedule is not None else 0
            count_by_resource_type["notification"] += len(notifications)

        print(f"Finished importing from {source} to {destination}.")
        table = Table(title="Import transformation-cli Summary")
        table.add_column("Resource Type", justify="right", style="cyan")
        table.add_column("Count", justify="right", style="magenta")
        for resource_type, count in count_by_resource_type.items():
            table.add_row(resource_type, str(count))
        print(table)

        if clean:
            if source.is_dir():
                print(f"Cleaning up source directory {source}.")
                safe_rmtree(source)
            else:
                print(f"Cleaning up source file {source}.")
                source.unlink()

    def _load_file(self, yaml_file: Path) -> dict[str, Any] | None:
        content = read_yaml_file(yaml_file, expected_output="dict")
        required_keys = {"externalId", "name", "destination", "query"}
        if missing_keys := required_keys - content.keys():
            self.warn(
                LowSeverityWarning(
                    f"Missing required keys {missing_keys} in {yaml_file}. Likely not a Transformation manifest. Skipping."
                )
            )
            return None
        return content

    def _convert_transformation(
        self, transformation: dict[str, Any], source_file: Path
    ) -> tuple[dict[str, Any], Path | None]:
        if "shared" in transformation:
            transformation["isPublic"] = transformation.pop("shared")
        if "action" in transformation:
            transformation["conflictMode"] = transformation.pop("action")
        if "ignoreNullFields" not in transformation:
            # This is required by the API, but the transformation-cli sets it to true by default.
            transformation["ignoreNullFields"] = True
        if isinstance(transformation["destination"], str):
            # The API expects the destination to be a dict with a type key, while the transformation-cli
            # also allows a string.
            transformation["destination"] = {"type": transformation["destination"]}

        source_query_path: Path | None = None
        if isinstance(transformation["query"], dict):
            query = transformation.pop("query")
            if "file" in query:
                source_query_path = Path(query.pop("file"))

        if "dataSetId" in transformation:
            if "dataSetExternalId" in transformation:
                self.warn(LowSeverityWarning(f"Both dataSetId and dataSetExternalId are present in {source_file}."))
            else:
                data_set_external_id = self._lookup_dataset(transformation.pop("dataSetId"))
                if data_set_external_id is None:
                    self.warn(
                        LowSeverityWarning(f"Failed to find DataSet with id {transformation['dataSetId']} in CDF.")
                    )
                else:
                    transformation["dataSetExternalId"] = data_set_external_id

        if "authentication" in transformation:
            authentication = transformation["authentication"]
            if not isinstance(authentication, dict):
                self.warn(LowSeverityWarning(f"Invalid authentication format in {source_file}."))
            else:
                if "tokenUrl" in authentication:
                    authentication["tokenUri"] = authentication.pop("tokenUrl")

                if "read" in authentication or "write" in authentication:
                    # Read or Write in authentication.
                    transformation.pop("authentication")

                read = authentication.pop("read", None)
                if read and isinstance(read, dict):
                    if "tokenUrl" in read:
                        read["tokenUri"] = read.pop("tokenUrl")
                    transformation["sourceOidcCredentials"] = read

                write = authentication.pop("write", None)
                if write and isinstance(write, dict):
                    if "tokenUrl" in write:
                        write["tokenUri"] = write.pop("tokenUrl")
                    transformation["destinationOidcCredentials"] = write
        return transformation, source_query_path

    def _convert_notifications(
        self, transformation: dict[str, Any], external_id: str, source_file: Path
    ) -> list[dict[str, Any]]:
        notifications = []
        if "notifications" in transformation:
            notifications_raw = transformation.pop("notifications")
            if isinstance(notifications_raw, list) and all(isinstance(n, str) for n in notifications_raw):
                notifications = [
                    {"destination": email, "transformationExternalId": external_id} for email in notifications_raw
                ]
            else:
                self.warn(LowSeverityWarning(f"Invalid notifications format in {source_file}."))
        return notifications

    def _convert_schedule(
        self, transformation: dict[str, Any], external_id: str, source_file: Path
    ) -> dict[str, Any] | None:
        schedule: dict[str, Any] | None = None
        if "schedule" in transformation:
            schedule_raw = transformation.pop("schedule")
            if isinstance(schedule_raw, str):
                schedule = {"interval": schedule_raw}
            elif isinstance(schedule_raw, dict):
                schedule = schedule_raw
            else:
                self.warn(LowSeverityWarning(f"Invalid schedule format in {source_file}."))
            if isinstance(schedule, dict):
                schedule["externalId"] = external_id
        return schedule

    def _lookup_dataset(self, dataset_id: int) -> str | None:
        if dataset_id in self._dataset_external_id_by_id:
            return self._dataset_external_id_by_id[dataset_id]
        dataset = self.client.data_sets.retrieve(id=dataset_id)
        if dataset is None or dataset.external_id is None:
            return None
        self._dataset_external_id_by_id[dataset.id] = dataset.external_id
        return dataset.external_id

    @property
    def client(self) -> ToolkitClient:
        if self._client is None:
            if self._get_client is None:
                raise AuthenticationError(
                    "No Cognite Client available. Are you missing a .env file?"
                    "\nThis is required to look up dataset ids in "
                    "the transformation-cli manifest(s)."
                )
            self._client = self._get_client()
        return self._client



import io
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, cast, final

import pandas as pd
from cognite.client.data_classes import FileMetadataWrite

from cognite_toolkit._cdf_tk.client.data_classes.extendable_cognite_file import ExtendableCogniteFileApply
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawTable
from cognite_toolkit._cdf_tk.constants import BUILD_FOLDER_ENCODING
from cognite_toolkit._cdf_tk.protocols import (
    T_ResourceRequest,
    T_ResourceRequestList,
    T_ResourceResponse,
    T_ResourceResponseList,
)
from cognite_toolkit._cdf_tk.utils import read_yaml_content, safe_read
from cognite_toolkit._cdf_tk.utils.file import read_csv

from ._base_cruds import T_ID, DataCRUD, ResourceCRUD
from ._resource_cruds import CogniteFileCRUD, FileMetadataCRUD, RawTableCRUD, TimeSeriesCRUD

if TYPE_CHECKING:
    from cognite_toolkit._cdf_tk.data_classes import BuildEnvironment


@final
class DatapointsCRUD(DataCRUD):
    item_name = "datapoints"
    folder_name = "timeseries"
    kind = "Datapoints"
    filetypes = frozenset({"csv", "parquet"})
    dependencies = frozenset({TimeSeriesCRUD})
    _doc_url = "Time-series/operation/postMultiTimeSeriesDatapoints"

    @property
    def display_name(self) -> str:
        return "timeseries datapoints"

    def upload(self, state: "BuildEnvironment", dry_run: bool) -> Iterable[tuple[str, int]]:
        if self.folder_name not in state.built_resources:
            return

        resource_directories = state.built_resources[self.folder_name].get_resource_directories(self.folder_name)

        for resource_dir in resource_directories:
            for datafile in self._find_data_files(resource_dir):
                if datafile.suffix == ".csv":
                    # The replacement is used to ensure that we read exactly the same file on Windows and Linux
                    file_content = datafile.read_bytes().replace(b"\r\n", b"\n").decode("utf-8")
                    data = read_csv(io.StringIO(file_content), parse_dates=True, index_col=0)
                    data.index = pd.DatetimeIndex(data.index)
                elif datafile.suffix == ".parquet":
                    data = pd.read_parquet(datafile, engine="pyarrow")
                else:
                    raise ValueError(f"Unsupported file type {datafile.suffix} for {datafile.name}")
                timeseries_ids = list(data.columns)
                if len(timeseries_ids) == 1:
                    ts_str = timeseries_ids[0]
                elif len(timeseries_ids) <= 10:
                    ts_str = str(timeseries_ids)
                else:
                    ts_str = f"{len(timeseries_ids):,} timeseries"

                if data.empty:
                    yield (
                        f"Empty file {datafile.as_posix()!r}. No datapoints to inserted.",
                        0,
                    )
                    continue

                if dry_run:
                    yield (
                        f" Would insert '{len(data):,}x{len(data.columns):,}' datapoints from '{datafile!s}' into {ts_str}",
                        len(data) * len(data.columns),
                    )
                else:
                    self.client.time_series.data.insert_dataframe(data)

                    yield (
                        f" Inserted '{len(data):,}x{len(data.columns):,}' datapoints from '{datafile!s}' into {ts_str}",
                        len(data) * len(data.columns),
                    )


@final
class FileCRUD(DataCRUD):
    item_name = "file contents"
    folder_name = "files"
    kind = "File"
    filetypes = frozenset()
    exclude_filetype: frozenset[str] = frozenset({})
    filename_pattern = (
        # Exclude FileMetadata and CogniteFile
        r"(?i)^(?!.*(?:FileMetadata|CogniteFile)$).*$"
    )
    dependencies = frozenset({FileMetadataCRUD, CogniteFileCRUD})
    _doc_url = "Files/operation/initFileUpload"

    @property
    def display_name(self) -> str:
        return "file content"

    def upload(self, state: "BuildEnvironment", dry_run: bool) -> Iterable[tuple[str, int]]:
        if self.folder_name not in state.built_resources:
            return

        for resource in state.built_resources[self.folder_name]:
            if resource.destination is None:
                continue

            if result := {
                FileMetadataCRUD.kind: (FileMetadataCRUD, "external_id"),
                CogniteFileCRUD.kind: (CogniteFileCRUD, "instance_id"),
            }.get(resource.kind):
                loader_cls, id_name = result
                meta: FileMetadataWrite | ExtendableCogniteFileApply = self._read_metadata(
                    resource.destination,
                    loader_cls,  # type: ignore[arg-type]
                    resource.identifier,
                )
                if meta.name is None:
                    continue
                datafile = resource.source.path.parent / meta.name
                if not datafile.exists():
                    continue

                if dry_run:
                    yield f" Would upload file '{datafile!s}' to file with {id_name}={resource.identifier!r}", 1
                else:
                    self.client.files.upload_content(path=str(datafile), **{id_name: resource.identifier})
                    yield f" Uploaded file '{datafile!s}' to file with {id_name}={resource.identifier!r}", 1

    @staticmethod
    def _read_metadata(
        destination: Path,
        loader: type[
            ResourceCRUD[T_ID, T_ResourceRequest, T_ResourceResponse, T_ResourceRequestList, T_ResourceResponseList]
        ],
        identifier: T_ID,
    ) -> T_ResourceRequest:
        built_content = read_yaml_content(safe_read(destination, encoding=BUILD_FOLDER_ENCODING))
        if isinstance(built_content, dict):
            return loader.resource_write_cls._load(built_content)
        elif isinstance(built_content, list):
            try:
                return next(m for m in loader.list_write_cls.load(built_content) if loader.get_id(m) == identifier)
            except StopIteration:
                raise RuntimeError(f"Missing metadata for {destination.as_posix()}")

        raise RuntimeError(f"Unexpected content type {type(built_content)} in {destination.as_posix()}")


@final
class RawFileCRUD(DataCRUD):
    item_name = "rows"
    folder_name = "raw"
    filetypes = frozenset({"csv", "parquet"})
    kind = "Raw"
    dependencies = frozenset({RawTableCRUD})
    _doc_url = "Raw/operation/postRows"

    @property
    def display_name(self) -> str:
        return "raw rows"

    def upload(self, state: "BuildEnvironment", dry_run: bool) -> Iterable[tuple[str, int]]:
        if self.folder_name not in state.built_resources:
            return

        for resource in state.built_resources[self.folder_name]:
            if resource.kind != RawTableCRUD.kind:
                continue
            table = cast(RawTable, resource.identifier)
            datafile = next(
                (
                    resource.source.path.with_suffix(f".{file_type}")
                    for file_type in self.filetypes
                    if (resource.source.path.with_suffix(f".{file_type}").exists())
                ),
                None,
            )
            if datafile is None:
                # No adjacent data file found
                continue

            if datafile.suffix == ".csv":
                # The replacement is used to ensure that we read exactly the same file on Windows and Linux
                file_content = datafile.read_bytes().replace(b"\r\n", b"\n").decode("utf-8")
                data = read_csv(io.StringIO(file_content)).astype(object).fillna("")
                if not data.columns.empty and data.columns[0] == "key":
                    print(f"Setting index to 'key' for {datafile.name}")
                    data.set_index("key", inplace=True)
            elif datafile.suffix == ".parquet":
                data = pd.read_parquet(datafile, engine="pyarrow")
            else:
                raise ValueError(f"Unsupported file type {datafile.suffix} for {datafile.name}")

            if data.empty:
                yield (
                    f"Empty file {datafile.as_posix()!r}. No rows to insert into {table!r}.",
                    0,
                )
                continue

            if dry_run:
                yield (
                    (
                        f" Would insert {len(data):,} rows of {len(data.columns):,} columns from '{datafile!s}' "
                        f"into {table!r}."
                    ),
                    len(data),
                )
                continue

            self.client.raw.rows.insert_dataframe(
                db_name=table.db_name, table_name=table.table_name, dataframe=data, ensure_parent=False
            )
            yield (
                (f" Inserted {len(data):,} rows of {len(data.columns):,} columns from '{datafile!s}' into {table!r}."),
                len(data),
            )



from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes import FunctionScheduleYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


class TestFunctionScheduleYAML:
    @pytest.mark.parametrize("data", list(find_resources("schedule", "functions")))
    def test_load_valid_function_schedule(self, data: dict[str, object]) -> None:
        loaded = FunctionScheduleYAML.model_validate(data)

        dumped = loaded.model_dump(exclude_unset=True, by_alias=True)
        assert "authentication" in dumped
        # Secret is not dumped as per design, so we add it back for comparison
        dumped["authentication"]["clientSecret"] = data["authentication"]["clientSecret"]
        assert dumped == data

    @pytest.mark.parametrize(
        "data, expected_errors",
        [
            pytest.param(
                {"functionExternalId": "fn_first_function", "name": "daily-8am-utc"},
                ["Missing required field: 'cronExpression'"],
                id="missing_required_field",
            )
        ],
    )
    def test_load_invalid_function_schedule(self, data: dict[str, object], expected_errors: list[str]) -> None:
        warnings = validate_resource_yaml_pydantic(data, FunctionScheduleYAML, source_file=Path("test.yaml"))
        assert len(warnings) == 1, "Only one warning should be raised"
        warning = warnings[0]
        assert isinstance(warning, ResourceFormatWarning), (
            f"Warning should be of type ResourceFormatWarning but got {type(warning)}"
        )
        assert list(warning.errors) == expected_errors, (
            f"Warning should contain the expected error {expected_errors} but got {list(warning.errors)}"
        )

    @pytest.mark.parametrize(
        "data",
        [
            pytest.param(
                {
                    "functionExternalId": "my_function",
                    "name": "my_schedule",
                    "cronExpression": "0 8 * * *",
                    "data": {"some": {"nested": "data"}},
                },
                id="Schedule with nested data",
            )
        ],
    )
    def test_load_valid_function(self, data: dict[str, object]) -> None:
        warnings = validate_resource_yaml_pydantic(data, FunctionScheduleYAML, source_file=Path("test.yaml"))

        assert len(warnings) == 0, "No warnings should be raised for valid function schedule"
